"""
This module is necessary since the docker python library does not yet provide built-in support
for nvidia-docker tooling. If some incarnation of this feature[1] makes its way into a release
of docker-py, this module can go away. Until then, the ideas in that PR have been copied and
lightly modified here.

[1] https://github.com/docker/docker-py/pull/1560
"""

import os
import re
import requests

from docker.api.client import APIClient
from docker.client import DockerClient
from docker.errors import DockerException
from docker.utils.utils import parse_devices

NVIDIA_DEFAULT_HOST = 'localhost'
NVIDIA_DEFAULT_PORT = 3476


def get_nvidia_docker_endpoint():
    host = os.environ.get('NV_HOST', 'http://%s:%s' % (NVIDIA_DEFAULT_HOST, NVIDIA_DEFAULT_PORT))
    return host + '/docker/cli/json'


def get_nvidia_configuration():
    url = get_nvidia_docker_endpoint()
    try:
        return requests.get(url).json()
    except requests.exceptions.ConnectionError:
        url = get_nvidia_docker_endpoint()
        raise NvidiaConnectionError((
            'Couldn\'t connect to nvidia-driver-plugin at {url} - is it running and accessible?.\n'
            'Try: "curl {url}" or "systemctl start nvidia-docker"').format(url=url), url)


def is_nvidia_image(api, image):
    labels = api.inspect_image(image).get('Config', {}).get('Labels')
    return bool(labels and labels.get('com.nvidia.volumes.needed') == 'nvidia_driver')


def add_nvidia_docker_to_config(container_config):
    if not container_config.get('HostConfig', None):
        container_config['HostConfig'] = {}

    nvidia_config = get_nvidia_configuration()

    # Setup the Volumes
    container_config['HostConfig'].setdefault('VolumeDriver', nvidia_config['VolumeDriver'])
    container_config['HostConfig'].setdefault('Binds', [])
    container_config['HostConfig']['Binds'].extend(nvidia_config['Volumes'])

    # Get nvidia control devices
    devices = container_config['HostConfig'].get('Devices', [])
    # suport both '0 1' and '0, 1' formats, just like nvidia-docker
    gpu_isolation = os.getenv('NV_GPU', '').replace(',', ' ').split()
    pattern = re.compile(r'/nvidia([0-9]+)$')
    for device in nvidia_config['Devices']:
        if gpu_isolation:
            card_number = pattern.search(device)
            if card_number and card_number.group(1) not in gpu_isolation:
                continue
        devices.extend(parse_devices([device]))

    container_config['HostConfig']['Devices'] = devices


class NvidiaDockerClient(DockerClient):
    def __init__(self, *args, **kwargs):
        self.api = NvidiaAPIClient(*args, **kwargs)


class NvidiaAPIClient(APIClient):
    def create_container_config(self, image, *args, **kwargs):
        container_config = (
            super(NvidiaAPIClient, self).create_container_config(image, *args, **kwargs))

        if is_nvidia_image(self, image):
            add_nvidia_docker_to_config(container_config)

        return container_config


class NvidiaConnectionError(DockerException):
    def __init__(self, msg, nvidia_url):
        self.msg = msg
        self.url = nvidia_url



import csv
import six

output = six.StringIO()
writer = csv.DictWriter(output, input['fields'], delimiter='\t')
writer.writerow({d: d for d in input['fields']})
for d in input['rows']:
    writer.writerow(d)
output = output.getvalue()



"""This module defines the PortList spec."""
import six

from .spec import SpecMixin
from .port import Port


class PortList(SpecMixin, list):
    """A list that only accepts port specs.

    This class is extended to behave like a read-only dictionary, where
    the keys are the port names.

    >>> l = PortList()
    >>> l.json()
    '[]'

    Ports can be added as instances or dictionaries

    >>> l.append(Port(name='z'))
    >>> l.append({'name': 'a', 'type': 'image', 'format': 'png'})

    Normal list methods are supported

    >>> l[1] = '{"name": "b"}'
    >>> l.insert(1, {"name": "c"})
    >>> del l[1]
    >>> str(l)
    '[{"name": "z"}, {"name": "b"}]'

    Port lists have keys and values methods like dicts

    >>> l.keys()[0]
    'z'

    Ports can be referenced by either their index in the list or by name

    >>> l[0] is l['z']
    True
    >>> 'z' in l
    True

    Ports can be modified after they are added

    >>> l[0].name = 'y'
    >>> l.json()
    '[{"name": "y"}, {"name": "b"}]'

    Several validation checks are performed

    >>> l[0].name = 'b'
    Traceback (most recent call last):
        ...
    ValueError: Duplicate keys detected
    >>> l[0].name = 'a'
    >>> l[0].format = 'png'
    Traceback (most recent call last):
        ...
    ValueError: Unknown format "python.png"
    >>> str(l)
    '[{"name": "a"}, {"name": "b"}]'
    """

    def __init__(self, value=None):
        """Create an empty list and optionally populate with an iterable."""
        SpecMixin.__init__(self)
        # list.__init__(self)
        if value is not None:
            for v in value:
                self.append(v)

    def __contains__(self, value):
        """Extend to allow dict like behavior."""
        if isinstance(value, six.string_types):
            try:
                value = self.keys().index(value)
            except ValueError:
                return False
            return True
        return list.__contains__(self, value)

    def __getitem__(self, index):
        """Extend to allow indexing by port name."""
        if isinstance(index, six.string_types):
            # find the index of the port with the given name
            index = self.keys().index(index)
        return list.__getitem__(self, index)

    def __setitem__(self, index, value):
        """Replace an item in the list if possible."""
        value = self._make_port(value)
        self._can_insert(value)
        list.__setitem__(self, index, value)

    def __delitem__(self, index):
        """Remove an item from the port list, freeing handlers."""
        self[index].remove_validation_check('PortList.name')
        list.__delitem__(self, index)

    def insert(self, index, value):
        """Add an item before the given index if possible."""
        value = self._make_port(value)
        self._can_insert(value)
        list.insert(self, index, value)

    def append(self, value):
        """Append an item if possible."""
        self.insert(len(self), value)

    def _assert_no_duplicates(self):
        """Raise a ValueError if there are duplicate port names."""
        if len(set(self.keys())) != len(self):
            raise ValueError('Duplicate keys detected')

    def _port_changed_handler(self):
        """Return a function that handles port spec changes.

        This function assures that no port names are changed resulting
        in a duplicate key.
        """
        def handler(port, key=None, oldvalue=None, newvalue=None):
            if key in ('name', None):
                self._assert_no_duplicates()
        return handler

    def _can_insert(self, port):
        """Raise a ValueError if the port cannot be inserted.

        This method also binds validation handlers to the port specs.
        """
        if port.name in self.keys():
            raise ValueError('A port named "%s" already exists' % port.name)
        port.add_validation_check('PortList.name', self._port_changed_handler())

    def check(self):
        """Check that the port list is valid."""
        self._assert_no_duplicates()
        for port in self:
            port.check()

    def _make_port(self, port):
        """Coererce the arguement into a port spec."""
        return Port(port)

    def keys(self):
        """Return a list of port names."""
        return [port.name for port in self]

    def values(self):
        """Return a list of ports."""
        return list(self)



from bson.json_util import dumps
from bson.objectid import ObjectId

output = []
node_oids = {}

# There is no clearly defined notion of undirected graphs in clique.json
# It simply checks for edges going in either direction, so we ignore
# whether it is an nx.DiGraph or just nx.Graph

for (node, data) in input.nodes(data=True):
    node_oids[node] = ObjectId()

    clique_node = {
        '_id': node_oids[node],
        'type': 'node'
    }

    if data:
        clique_node['data'] = data

    output.append(clique_node)

for (u, v, edge_data) in input.edges(data=True):
    clique_edge = {
        '_id': ObjectId(),
        'source': node_oids[u],
        'target': node_oids[v],
        'type': 'link'
    }

    if 'data' in edge_data:
        clique_edge['data'] = edge_data['data']

    output.append(clique_edge)

output = dumps(output)



import os
from six.moves import configparser
import girder_worker

accept_content = ['json', 'pickle', 'yaml', 'girder_io']

broker_url = os.environ.get('GIRDER_WORKER_BROKER', None) or \
    girder_worker.config.get('celery', 'broker')

try:
    result_backend = os.environ.get('GIRDER_WORKER_BACKEND', None) or \
        girder_worker.config.get('celery', 'backend')
except configparser.NoOptionError:
    result_backend = broker_url



"""Tests for core spec objects."""
import unittest
from unittest import TestCase
from girder_worker.core import specs, run
from girder_worker.core.specs.utils import spec_class_generator, to_frozenset


class TestSpec(TestCase):
    """Tests edge cases of the base spec."""

    def test_key_not_str(self):
        """Raise a TypeError for invalid keys."""
        s = specs.Spec()
        with self.assertRaises(TypeError):
            s.update({0: 'value'})

    def test_delete_required_exception(self):
        """Raise an exception during a key deletion."""
        def has_required(self, *a, **kw):
            assert 'required' in self

        s = specs.Spec(required=True)
        s.add_validation_check('required', has_required)
        s.required = False  # truthiness of the value doesn't matter

        with self.assertRaises(AssertionError):
            del s['required']


class TestPort(TestCase):
    """Tests edge cases of the port spec."""

    def test_port_fetch(self):
        """Test edge cases in Port.fetch not testing in doctests."""
        port = specs.Port(name='a', type='number', format='number')

        # matching formats
        self.assertEqual(2, port.fetch({'data': 2, 'format': 'number'}))

        # invalid data
        with self.assertRaises(specs.ValidationError):
            port.fetch({'data': '2', 'format': 'number'})

        # non-matching formats
        self.assertEqual(2, port.fetch({'data': '2', 'format': 'json'}))

        # not convertible
        port.auto_convert = False
        with self.assertRaises(Exception):
            port.fetch({'data': '2', 'format': 'json'})

    def test_port_push(self):
        """Test edge cases in Port.push not testing in doctests."""
        port = specs.Port(name='a', type='number', format='number')

        # matching formats
        self.assertEqual(2, port.push({'data': 2, 'format': 'number'})['data'])

        # invalid data
        with self.assertRaises(specs.ValidationError):
            port.push({'data': '2', 'format': 'number'})

        # non-matching formats
        self.assertEqual(2, port.push({'data': '2', 'format': 'json'})['data'])

        # not convertible
        port.auto_convert = False
        with self.assertRaises(Exception):
            port.push({'data': '2', 'format': 'json'})

    def test_port_init(self):
        """Test edge cases in Port not testing in doctests."""
        p = specs.Port(name='a')
        with self.assertRaises(ValueError):
            p.type = 'notatype'


class TestTaskSpec(TestCase):
    """Tests edge cases of the anonymous task spec."""

    def test_task_inputs_outputs_equality(self):
        """
        Test input and output equality through specs, task __getitem__
        interface and task __getattr__ interface.
        """
        inputs = sorted([
            {'name': 'a', 'type': 'string', 'format': 'text'},
            {'name': 'b', 'type': 'number', 'format': 'number'},
            {'name': 'c', 'type': 'number', 'format': 'number'},
        ])
        outputs = sorted([
            {'name': 'd', 'type': 'string', 'format': 'text'}
        ])

        spec = {
            'inputs': inputs,
            'outputs': outputs,
            'script': "d = a + ':' + str(b + c)"}

        t = specs.TaskSpec(spec)

        self.assertEquals(sorted(spec['inputs']), inputs)
        self.assertEqual(sorted(spec['outputs']), outputs)

        self.assertEquals(sorted(t['inputs']), inputs)
        self.assertEqual(sorted(t['outputs']), outputs)

        self.assertEquals(sorted(t.inputs), inputs)
        self.assertEquals(sorted(t.outputs), outputs)


class TestTask(TestCase):
    """Tests edge cases of the task spec."""

    def setUp(self):
        self.inputs = sorted([
            {'name': 'a', 'type': 'string', 'format': 'text'},
            {'name': 'b', 'type': 'number', 'format': 'number'},
            {'name': 'c', 'type': 'number', 'format': 'number'},
        ])
        self.outputs = sorted([
            {'name': 'd', 'type': 'string', 'format': 'text'}
        ])

        self.spec = {
            'script': "d = a + ':' + str(b + c)",
            'mode': 'python'
        }

    def test_mutable_mapping_interface(self):
        class TempTask(specs.Task):
            __inputs__ = specs.PortList(self.inputs)
            __outputs__ = specs.PortList(self.outputs)

        t = specs.Task(self.spec)

        self.assertEquals(t['script'], t.get('script'))
        self.assertEquals('default', t.get('non_present_key', 'default'))

    def test_class_level_set_of_inputs_outputs(self):

        """Test task input/output attributes are set from class vars"""

        class TempTask(specs.Task):
            __inputs__ = specs.PortList(self.inputs)
            __outputs__ = specs.PortList(self.outputs)

        t = specs.Task(self.spec)
        self.assertEquals(
            set(t.keys()), {'inputs', 'outputs', 'mode', 'script'})

        self.assertEquals(t['inputs'], specs.PortList())
        self.assertEquals(t['outputs'], specs.PortList())

        t2 = TempTask(self.spec)
        self.assertEquals(t2['inputs'], self.inputs)
        self.assertEquals(t2['outputs'], self.outputs)

    def test_read_only_attributes(self):

        """Raise exception if task input/output are assigned"""

        class TempTask(specs.Task):
            __inputs__ = specs.PortList(self.inputs)
            __outputs__ = specs.PortList(self.outputs)

        # Test if passed in in spec dict
        with self.assertRaises(specs.ReadOnlyAttributeException):
            spec = self.spec.copy()
            spec['inputs'] = specs.PortList()
            TempTask(spec)

        with self.assertRaises(specs.ReadOnlyAttributeException):
            spec = self.spec.copy()
            spec['outputs'] = specs.PortList()
            TempTask(spec)

        # Test if assigned after instatiation
        t = TempTask(self.spec)

        with self.assertRaises(specs.ReadOnlyAttributeException):
            t['inputs'] = specs.PortList()

        with self.assertRaises(specs.ReadOnlyAttributeException):
            t['outputs'] = specs.PortList()

        with self.assertRaises(specs.ReadOnlyAttributeException):
            t.inputs = specs.PortList()

        with self.assertRaises(specs.ReadOnlyAttributeException):
            t.outputs = specs.PortList()


class TestWorkflow(TestCase):

    def assertConsistent(self, system, ground, type_spec=None):
        """
        Assert that the system and ground are consistent i.e., they
        have the same order-variant dicts. We test raw dicts,
        and Spec dicts and if present type specific dicts.
        """

        # Test dicts
        self.assertEquals(to_frozenset(system),
                          to_frozenset(ground))
        # Test Specs
        self.assertEquals(to_frozenset(system),
                          to_frozenset([specs.Spec(d)
                                        for d in ground]))
        if type_spec is not None:
            # Type specific spec
            self.assertEquals(to_frozenset(system),
                              to_frozenset([type_spec(d)
                                            for d in ground]))

    def setUp(self):
        self.add = {
            'inputs': [{'name': 'a', 'type': 'number', 'format': 'number'},
                       {'name': 'b', 'type': 'number', 'format': 'number'}],
            'outputs': [{'name': 'c', 'type': 'number', 'format': 'number'}],
            'script': 'c = a + b',
            'mode': 'python'}

        self.add_three = {
            'inputs': [{'name': 'a', 'type': 'number', 'format': 'number'}],
            'outputs': [{'name': 'b', 'type': 'number', 'format': 'number'}],
            'mode': 'python',
            'script': 'b = a + 3'}

        self.add_two = {
            'inputs': [{'name': 'a', 'type': 'number', 'format': 'number'}],
            'outputs': [{'name': 'b', 'type': 'number', 'format': 'number'}],
            'mode': 'python',
            'script': 'b = a + 2'}

        self.multiply = {
            'inputs': [{'name': 'in1', 'type': 'number', 'format': 'number'},
                       {'name': 'in2', 'type': 'number', 'format': 'number'}],
            'outputs': [{'name': 'out', 'type': 'number', 'format': 'number'}],
            'mode': 'python',
            'script': 'out = in1 * in2'}

        self.workflow = {
            'mode': 'workflow',

            'inputs': [{'name': 'a3.a', 'type': 'number', 'format': 'number',
                        'default': {'format': 'number', 'data': 10}},
                       {'name': 'a2.a', 'type': 'number', 'format': 'number'}],

            'outputs': [{'name': 'out', 'type': 'number', 'format': 'number'}],

            'steps': [{'task': self.add_three, 'name': 'a3'},
                      {'task': self.add_two, 'name': 'a2'},
                      {'task': self.multiply, 'name': 'm'}],

            'connections': [
                {
                    'name': 'a3.a',
                    'input_step': 'a3',
                    'input': 'a'
                },
                {
                    'name': 'a2.a',
                    'input_step': 'a2',
                    'input': 'a'
                },
                {
                    'output_step': 'a3',
                    'output': 'b',
                    'input_step': 'm',
                    'input': 'in1'
                },
                {
                    'output_step': 'a2',
                    'output': 'b',
                    'input_step': 'm',
                    'input': 'in2'
                },
                {
                    'name': 'out',
                    'output_step': 'm',
                    'output': 'out'
                }
            ]
        }

    def test_spec_class_generator(self):
        """
        Instantiated classes from spec_class_generator should equal
        their spec.
        """
        for spec in [self.add, self.add_three, self.add_two, self.multiply]:
            cls = spec_class_generator('cls', spec)
            self.assertEqual(cls(), spec)

    def test_empty_workflow(self):
        """Empty Workflow objects should have certain properties"""
        wf = specs.Workflow()

        self.assertEquals(len(wf), 5)
        self.assertEquals(set(wf.keys()), set(['mode', 'steps',
                                               'connections', 'inputs',
                                               'outputs']))
        self.assertEquals(wf['mode'], 'workflow')
        self.assertEquals(wf['steps'], [])
        self.assertEquals(wf['connections'], [])
        self.assertEquals(wf['inputs'], [])
        self.assertEquals(wf['outputs'], [])

        self.assertEquals(wf.mode, 'workflow')
        self.assertEquals(wf.steps, [])
        self.assertEquals(wf.connections, [])
        self.assertEquals(wf.inputs, [])
        self.assertEquals(wf.outputs, [])

        with self.assertRaises(specs.ReadOnlyAttributeException):
            del wf['mode']

        with self.assertRaises(KeyError):
            del wf['foo']

        with self.assertRaises(specs.ReadOnlyAttributeException):
            wf['mode'] = 'foo'

        with self.assertRaises(KeyError):
            wf['foo']

        with self.assertRaises(KeyError):
            wf['foo'] = 'foo'

    def test_workflow_add_task_dict(self):
        """Adding task dicts should show up in Workflow.steps"""
        wf = specs.Workflow()

        wf.add_task(self.add, 'add')
        task_list = [{'name': 'add', 'task': self.add}]
        # Steps should be equal to a list of dicts
        self.assertEquals(wf['steps'], task_list)
        self.assertEquals(wf.steps, task_list)

        # Steps should be equal to a list of Spec() dicts
        self.assertEquals(wf['steps'], [specs.Spec(t) for t in task_list])
        self.assertEquals(wf.steps, [specs.Spec(t) for t in task_list])

        # Steps should be equal to a list of StepSpec() dicts
        self.assertEquals(wf['steps'], [specs.StepSpec(t) for t in task_list])
        self.assertEquals(wf.steps, [specs.StepSpec(t) for t in task_list])

        # No duplicate nodes
        with self.assertRaises(specs.DuplicateTaskException):
            wf.add_task(self.add, 'add')

        # Test with multiple tasks of the same type (different name)
        wf.add_task(self.add, 'add2')
        task_list += [{'name': 'add2', 'task': self.add}]

        self.assertEquals(to_frozenset(wf['steps']), to_frozenset(task_list))
        self.assertEquals(to_frozenset(wf.steps), to_frozenset(task_list))

        self.assertEquals(to_frozenset(wf['steps']),
                          to_frozenset([specs.Spec(t) for t in task_list]))
        self.assertEquals(to_frozenset(wf.steps),
                          to_frozenset([specs.Spec(t) for t in task_list]))

        # Steps should be equal to a list of StepSpec() dicts
        self.assertEquals(to_frozenset(wf['steps']),
                          to_frozenset([specs.StepSpec(t) for t in task_list]))
        self.assertEquals(to_frozenset(wf.steps),
                          to_frozenset([specs.StepSpec(t) for t in task_list]))

    def test_workflow_connect_tasks(self):
        """Verify connections is correct given a known task graph"""

        # * Test connect_tasks(t1, t2, {'output': input})
        inputs_ground = [{'format': 'number', 'name': 'a', 'type': 'number'}]
        outputs_ground = [{'format': 'number', 'name': 'b', 'type': 'number'}]
        steps_ground = [{'name': 'a2', 'task': self.add_two},
                        {'name': 'a3', 'task': self.add_three}]
        connections_ground = [{'input': 'a', 'input_step': 'a2', 'name': 'a'},
                              {'input': 'a', 'input_step': 'a3',
                               'output': 'b', 'output_step': 'a2'},
                              {'name': 'b', 'output': 'b', 'output_step': 'a3'}]

        wf = specs.Workflow()
        wf.add_task(self.add_two, 'a2')
        wf.add_task(self.add_three, 'a3')

        wf.connect_tasks('a2', 'a3', {'b': 'a'})

        self.assertConsistent(wf.inputs, inputs_ground, type_spec=specs.Port)
        self.assertConsistent(wf['inputs'], inputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.outputs, outputs_ground, type_spec=specs.Port)
        self.assertConsistent(
            wf['outputs'], outputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.steps, steps_ground)
        self.assertConsistent(wf['steps'], steps_ground)

        self.assertConsistent(wf.connections, connections_ground)
        self.assertConsistent(wf['connections'], connections_ground)

        #####
        # * Test connect_tasks(t1, t2, output='input')

        wf = specs.Workflow()
        wf.add_task(self.add_two, 'a2')
        wf.add_task(self.add_three, 'a3')

        wf.connect_tasks('a2', 'a3', b='a')

        self.assertConsistent(wf.inputs, inputs_ground, type_spec=specs.Port)
        self.assertConsistent(wf['inputs'], inputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.outputs, outputs_ground, type_spec=specs.Port)
        self.assertConsistent(
            wf['outputs'], outputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.steps, steps_ground)
        self.assertConsistent(wf['steps'], steps_ground)

        self.assertConsistent(wf.connections, connections_ground)
        self.assertConsistent(wf['connections'], connections_ground)

        #####
        # * Test connect_tasks((t1, t2, {'output', 'input'}),
        #                      (t3, t2, {'output', 'input'}))

        inputs_ground = [{'format': 'number', 'name': 'a', 'type': 'number'}]
        outputs_ground = [{'format': 'number', 'name': 'b', 'type': 'number'}]
        steps_ground = [{'name': 'a1', 'task': self.add_two},
                        {'name': 'a2', 'task': self.add_two},
                        {'name': 'a3', 'task': self.add_three}]
        connections_ground = [{'input': 'a', 'input_step': 'a1', 'name': 'a'},
                              {'input': 'a', 'input_step': 'a2',
                               'output': 'b', 'output_step': 'a1'},
                              {'input': 'a', 'input_step': 'a3',
                               'output': 'b', 'output_step': 'a2'},
                              {'name': 'b', 'output': 'b', 'output_step': 'a3'}]

        wf = specs.Workflow()

        wf.add_task(self.add_two, 'a1')
        wf.add_task(self.add_two, 'a2')
        wf.add_task(self.add_three, 'a3')

        wf.connect_tasks([('a1', 'a2', {'b': 'a'}),
                          ('a2', 'a3', {'b': 'a'})])

        self.assertConsistent(wf.inputs, inputs_ground, type_spec=specs.Port)
        self.assertConsistent(wf['inputs'], inputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.outputs, outputs_ground, type_spec=specs.Port)
        self.assertConsistent(
            wf['outputs'], outputs_ground, type_spec=specs.Port)

        self.assertConsistent(wf.steps, steps_ground)
        self.assertConsistent(wf['steps'], steps_ground)

        self.assertConsistent(wf.connections, connections_ground)
        self.assertConsistent(wf['connections'], connections_ground)

        #####
        # * Test connect_tasks with invalid input/output mapping
        # ** Doesn't exist
        #    TO IMPLEMENT

        # ** Isn't valid
        #    NOT CURRENTLY IMPLEMENTED/TESTED

        #####################################################################
        # * Edge Cases
        #
        #                         Task Graph 1
        #                        ==============
        #  Testing with a non-trival task graph including several edge cases
        #  This has multipe unconnected inputs that need to be propigated up
        #  To the top level inputs/outputs attributes on the workflow.
        #  Additionally the 'b' input on A2  will conflict with the 'b' input
        #  on A1.  This tests the conflict resolution strategy which is to
        #  prefix port names with task names (e.g.,  'a1.b'  and 'a2.b')
        #
        #
        #        +               +          +              +
        # {in1}  |        {in2}  |          | {a}          | {b}
        #        |               |          |              |
        #        +--^---------^--+          +-^---------^--+
        #           |         |               |         |
        #           |    M1   |               |   A1    |
        #           |         |               |         |
        #           +----+----+               +-----+---+
        #                |                          |
        #          {out} |                          |  {c}
        #                |                          |
        #                |       +-----------+      |
        #                |       |           |      |
        #                +------^+           +^-----+
        #                  {in1} |    M2     |  {in2}
        #                        |           |
        #                        +-----+-----+                  +
        #                              |                        |
        #                        {out} |                        |
        #                              |                        |
        #                              |    +------------+      |
        #                              |    |            |      |
        #                              +---^+            +^-----+
        #                              {a}  |    A2      |   {b}
        #                                   |            |
        #                                   +-----+------+
        #                                         |
        #                                         |  {c}
        #                                         |
        #                                         v
        #
        #################################################################
        wf = specs.Workflow()

        wf.add_task(self.multiply, 'm1')
        wf.add_task(self.multiply, 'm2')
        wf.add_task(self.add, 'a1')
        wf.add_task(self.add, 'a2')

        wf.connect_tasks('m1', 'm2', {'out': 'in1'})
        wf.connect_tasks('a1', 'm2', {'c': 'in2'})
        wf.connect_tasks('m2', 'a2', {'out': 'a'})

        connections_ground = [
            {'input': 'a', 'input_step': 'a1', 'name': 'a'},
            {'input': 'b', 'input_step': 'a1', 'name': 'a1.b'},
            {'input': 'b', 'input_step': 'a2', 'name': 'a2.b'},
            {'input': 'in1', 'input_step': 'm1', 'name': 'in1'},
            {'input': 'in2', 'input_step': 'm1', 'name': 'in2'},
            {'input': 'in2', 'input_step': 'm2',
             'output': 'c', 'output_step': 'a1'},
            {'input': 'in1', 'input_step': 'm2',
             'output': 'out', 'output_step': 'm1'},
            {'input': 'a', 'input_step': 'a2',
             'output': 'out', 'output_step': 'm2'},
            {'name': 'c', 'output': 'c', 'output_step': 'a2'}
        ]

        inputs_ground = [{'format': 'number', 'name': 'a', 'type': 'number'},
                         {'format': 'number', 'name': 'a1.b', 'type': 'number'},
                         {'format': 'number', 'name': 'a2.b', 'type': 'number'},
                         {'format': 'number', 'name': 'in1', 'type': 'number'},
                         {'format': 'number', 'name': 'in2', 'type': 'number'}]

        outputs_ground = [{'format': 'number', 'name': 'c', 'type': 'number'}]

        # Connections
        self.assertConsistent(wf.connections, connections_ground)
        self.assertConsistent(wf['connections'], connections_ground)

        # Inputs
        self.assertConsistent(wf.inputs, inputs_ground, type_spec=specs.Port)
        self.assertConsistent(wf['inputs'], inputs_ground, type_spec=specs.Port)

        # Outputs
        self.assertConsistent(wf.outputs, outputs_ground, type_spec=specs.Port)
        self.assertConsistent(
            wf['outputs'], outputs_ground, type_spec=specs.Port)

    def test_workflow(self):
        #                     Task Graph 2
        #                    ==============
        #  This is the same task graph from workflow_test.py with minor
        #  alterations (specifically the task names).
        #
        #               +                      +
        #            {a}|                      | {a}
        #               |                      |
        #            +--^----+              +--^----+
        #            |       |              |       |
        #            |  A3   |              |  A2   |
        #            |       |              |       |
        #            +---+---+              +---+---+
        #           {b}  |                      |  {b}
        #                |                      |
        #                |      +---------+     |
        #                |      |         |     |
        #                +-----^+    M    +^----+
        #                 {in1} |         | {in2}
        #                       +----+----+
        #                            |
        #                            |
        #                            | {out}
        #                            |
        #                            v

        wf = specs.Workflow()

        wf.add_task(self.add_two, 'a2')
        wf.add_task(self.add_three, 'a3')
        wf.add_task(self.multiply, 'm')

        wf.connect_tasks('a3', 'm', {'b': 'in1'})
        wf.connect_tasks('a2', 'm', {'b': 'in2'})

        # Add default as defined in self.workflow
        wf.set_default('a3.a', {'format': 'number', 'data': 10})

        # Asset that the components are equal and consistent
        self.assertConsistent(wf.inputs, self.workflow['inputs'],
                              type_spec=specs.Port)
        self.assertConsistent(wf['inputs'], self.workflow['inputs'],
                              type_spec=specs.Port)

        self.assertConsistent(wf.outputs, self.workflow['outputs'],
                              type_spec=specs.Port)
        self.assertConsistent(wf['outputs'], self.workflow['outputs'],
                              type_spec=specs.Port)

        self.assertConsistent(wf.steps, self.workflow['steps'],
                              type_spec=specs.StepSpec)
        self.assertConsistent(wf['steps'], self.workflow['steps'],
                              type_spec=specs.StepSpec)

        self.assertConsistent(wf.connections, self.workflow['connections'],
                              type_spec=specs.ConnectionSpec)
        self.assertConsistent(wf['connections'], self.workflow['connections'],
                              type_spec=specs.ConnectionSpec)

        # Assert the equality of the Workflow object and the workflow dict
        self.assertEquals(wf, self.workflow)

        inputs = {'a2.a': {'format': 'json', 'data': '1'},
                  'a3.a': {'format': 'number', 'data': 2}}

        ground = run(self.workflow, inputs=inputs)
        system = run(wf, inputs=inputs)

        self.assertEquals(system, ground)

    def test_workflow_with_generated_task_classes(self):
        wf = specs.Workflow()

        AddTwo = spec_class_generator('AddTwo', self.add_two)
        AddThree = spec_class_generator('AddThree', self.add_three)
        Multiply = spec_class_generator('Multiply', self.multiply)

        wf.add_task(AddTwo(), 'a2')
        wf.add_task(AddThree(), 'a3')
        wf.add_task(Multiply(), 'm')

        wf.connect_tasks('a3', 'm', {'b': 'in1'})
        wf.connect_tasks('a2', 'm', {'b': 'in2'})

        # Add default as defined in self.workflow
        wf.set_default('a3.a', {'format': 'number', 'data': 10})

        self.assertEquals(wf, self.workflow)
        inputs = {'a2.a': {'format': 'json', 'data': '1'},
                  'a3.a': {'format': 'number', 'data': 2}}

        ground = run(self.workflow, inputs=inputs)
        system = run(wf, inputs=inputs)

        self.assertEquals(system, ground)

    def test_workflow_with_task_classes(self):

        class AddTwo(specs.Task):
            __inputs__ = specs.PortList([
                {'name': 'a', 'type': 'number', 'format': 'number'}
            ])
            __outputs__ = specs.PortList([
                {'name': 'b', 'type': 'number', 'format': 'number'}
            ])

            def __init__(self, spec=None, **kw):
                super(AddTwo, self).__init__(spec, **kw)
                self.mode = 'python'
                self.script = 'b = a + 2'

        class AddThree(specs.Task):
            __inputs__ = specs.PortList([
                {'name': 'a', 'type': 'number', 'format': 'number'}
            ])
            __outputs__ = specs.PortList([
                {'name': 'b', 'type': 'number', 'format': 'number'}
            ])

            def __init__(self, spec=None, **kw):
                super(AddThree, self).__init__(spec, **kw)
                self.mode = 'python'
                self.script = 'b = a + 3'

        class Multiply(specs.Task):
            __inputs__ = specs.PortList([
                {'name': 'in1', 'type': 'number', 'format': 'number'},
                {'name': 'in2', 'type': 'number', 'format': 'number'}
            ])
            __outputs__ = specs.PortList([
                {'name': 'out', 'type': 'number', 'format': 'number'}
            ])

            def __init__(self, spec=None, **kw):
                super(Multiply, self).__init__(spec, **kw)
                self.mode = 'python'
                self.script = 'out = in1 * in2'

        wf = specs.Workflow()

        wf.add_task(AddTwo(), 'a2')
        wf.add_task(AddThree(), 'a3')
        wf.add_task(Multiply(), 'm')

        wf.connect_tasks('a3', 'm', {'b': 'in1'})
        wf.connect_tasks('a2', 'm', {'b': 'in2'})

        # Add default as defined in self.workflow
        wf.set_default('a3.a', {'format': 'number', 'data': 10})

        self.assertEquals(wf, self.workflow)
        inputs = {'a2.a': {'format': 'json', 'data': '1'},
                  'a3.a': {'format': 'number', 'data': 2}}

        ground = run(self.workflow, inputs=inputs)
        system = run(wf, inputs=inputs)

        self.assertEquals(system, ground)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Tests for the "Pickle" data type."""
import unittest
import tempfile
import six

from girder_worker.tasks import run
from girder_worker.plugins.types import convert


class TestPickle(unittest.TestCase):
    """Tests for the "Pickle" data type."""

    def round_trip(self, obj):
        """Convert an object to base64 and back returning the new object."""
        b64 = convert(
            'python',
            {'format': 'object', 'data': obj},
            {'format': 'pickle.base64'}
        )['data']
        newobj = convert(
            'python',
            {'format': 'pickle.base64', 'data': b64},
            {'format': 'object'}
        )
        return newobj['data']

    def assertRoundTrip(self, obj, descr):
        """Assert an object is unchanged in format conversion round trip."""
        self.assertEqual(obj, self.round_trip(obj), descr)

    def test_pickle_basic(self):
        """Test pickling basic Python types."""
        self.assertRoundTrip(0, 'int')
        self.assertRoundTrip(1.0, 'float')
        self.assertRoundTrip('a', 'str')
        self.assertRoundTrip(u'\u03c0', 'unicode')
        self.assertRoundTrip('\xff\xff', 'bytes')
        self.assertRoundTrip((0, 'a'), 'tuple')
        self.assertRoundTrip([0, 'a'], 'list')
        self.assertRoundTrip({'a': 0, 'b': None}, 'dict')
        self.assertRoundTrip(
            {'a': {'c': [1, 2, {'key': ('value',)}], 'b': None}},
            'nested dict'
        )

    def test_pickle_error(self):
        """Make sure an exception is raised for non-pickleable types."""
        with self.assertRaises(Exception):
            self.round_trip(tempfile.TemporaryFile())

        with self.assertRaises(Exception):
            self.round_trip(lambda x: x)

    def run_basic_task(self, inputs):
        """Run a basic task with pickle types."""
        script = """
c = a * b
d = b + b
"""
        task = {
            'inputs': [
                {'name': 'a', 'type': 'python', 'format': 'object'},
                {'name': 'b', 'type': 'python', 'format': 'object'}
            ],
            'outputs': [
                {'name': 'c', 'type': 'python', 'format': 'object'},
                {'name': 'd', 'type': 'python', 'format': 'object'}
            ],
            'script': script,
            'mode': 'python'
        }
        d = run(task, inputs=inputs, outputs={'format': 'object'})
        for k, v in six.iteritems(d):
            if isinstance(v, dict):
                d[k] = v.get('data')
        return d

    def test_basic_task(self):
        """Run a task without conversion."""
        outputs = self.run_basic_task({
            'a': {'format': 'object', 'data': '*'},
            'b': {'format': 'object', 'data': 4}}
        )

        self.assertEqual(outputs.get('c'), '****')
        self.assertEqual(outputs.get('d'), 8)

        outputs = self.run_basic_task({
            'a': {'format': 'object', 'data': 1.0},
            'b': {'format': 'object', 'data': 4}}
        )

        self.assertEqual(outputs.get('c'), 4.0)
        self.assertEqual(outputs.get('d'), 8)

    def test_inputs_from_file(self):
        """Run a task with base64 inputs in a file."""
        a = tempfile.NamedTemporaryFile()
        b = tempfile.NamedTemporaryFile()

        convert(
            'python',
            {'format': 'object', 'data': (0, 1)},
            {'format': 'pickle.base64', 'mode': 'local', 'path': a.name}
        )

        convert(
            'python',
            {'format': 'object', 'data': 2},
            {'format': 'pickle.base64', 'mode': 'local', 'path': b.name}
        )

        outputs = self.run_basic_task({
            'a': {'format': 'pickle.base64', 'mode': 'local', 'path': a.name},
            'b': {'format': 'pickle.base64', 'mode': 'local', 'path': b.name}
        })

        self.assertEqual(outputs.get('c'), (0, 1, 0, 1))
        self.assertEqual(outputs.get('d'), 4)



from girder_worker.tasks import run
from girder_worker.plugins.types import convert
from girder_worker.core import load
import os
import tempfile
import unittest
import collections
import json
import math


class TestTable(unittest.TestCase):

    def setUp(self):
        self.analysis = {
            'name': 'append_tables',
            'inputs': [
                {'name': 'a', 'type': 'table', 'format': 'rows'},
                {'name': 'b', 'type': 'table', 'format': 'rows'}
            ],
            'outputs': [{'name': 'c', 'type': 'table', 'format': 'rows'}],
            'script': ("c = {'fields': a['fields'], 'rows': a['rows'] +"
                       " b['rows']}"),
            'mode': 'python'
        }
        self.analysis_r = {
            'name': 'copy_table',
            'inputs': [
                {'name': 'a', 'type': 'table', 'format': 'r.dataframe'}
            ],
            'outputs': [
                {'name': 'b', 'type': 'table', 'format': 'r.dataframe'}
            ],
            'script': 'b <- a',
            'mode': 'r'
        }
        self.test_input = {
            'a': {
                'format': 'rows.json',
                'data': ('{"fields": ["aa", "bb"], '
                         '"rows": [{"aa": 1, "bb": 2}]}')
            },
            'b': {
                'format': 'rows.json',
                'data': ('{"fields": ["aa", "bb"],'
                         '"rows": [{"aa": 3, "bb": 4}]}')
            }
        }
        import pymongo
        import bson
        self.db = pymongo.MongoClient('mongodb://localhost')['test']
        self.db['a'].drop()
        self.aobj = collections.OrderedDict([
            ('_id', bson.ObjectId('530cdb1add29bc5628984303')),
            ('bar', 2.0), ('foo', 1.0)])
        self.db['a'].insert(self.aobj)
        self.db['b'].drop()
        self.bobj = collections.OrderedDict([
            ('_id', bson.ObjectId('530cdb23dd29bc5628984304')),
            ('bar', 4.0), ('foo', 3.0)])
        self.db['b'].insert(self.bobj)

        # Change dir for loading analyses
        self.prevdir = os.getcwd()
        cur_path = os.path.dirname(os.path.realpath(__file__))
        os.chdir(cur_path)
        self.analysis_path = os.path.join(cur_path, '..', 'analysis')

    def tearDown(self):
        os.chdir(self.prevdir)

    def test_json(self):
        outputs = run(
            self.analysis,
            inputs=self.test_input,
            outputs={
                'c': {'format': 'rows.json'}
            })
        self.assertEqual(outputs['c']['format'], 'rows.json')
        self.assertEqual(
            outputs['c']['data'],
            '{"fields": ["aa", "bb"], '
            '"rows": [{"aa": 1, "bb": 2}, {"aa": 3, "bb": 4}]}')

    def test_bson(self):
        import pymongo
        outputs = run(
            self.analysis,
            inputs={
                'a': {
                    'format': 'objectlist.bson',
                    'mode': 'mongodb',
                    'db': 'test',
                    'collection': 'a'
                },
                'b': {
                    'format': 'objectlist.bson',
                    'mode': 'mongodb',
                    'db': 'test',
                    'collection': 'b'
                }
            },
            outputs={
                'c': {
                    'format': 'objectlist.bson',
                    'mode': 'mongodb',
                    'db': 'test',
                    'collection': 'temp'
                }
            })
        self.assertEqual(outputs['c']['format'], 'objectlist.bson')
        coll = pymongo.MongoClient('mongodb://localhost')['test']['temp']
        self.assertEqual([d for d in coll.find()], [self.aobj, self.bobj])

    def test_file(self):
        tmp = tempfile.mktemp()
        outputs = run(
            self.analysis,
            inputs=self.test_input,
            outputs={
                'c': {'format': 'csv', 'path': tmp, 'mode': 'local'}
            }
        )
        with open(tmp, 'r') as fp:
            output = fp.read()
        os.remove(tmp)
        self.assertEqual(outputs['c']['format'], 'csv')
        self.assertEqual(output.splitlines(), ['aa,bb', '1,2', '3,4'])

    def test_csv(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'csv', 'data': 'a,b,c\n1,2,3'},
                'b': {'format': 'csv', 'data': 'a,b,c\n4,5,6'}
            },
            outputs={
                'c': {'format': 'csv'}
            })
        self.assertEqual(outputs['c']['format'], 'csv')
        self.assertEqual(outputs['c']['data'].splitlines(),
                         ['a,b,c', '1,2,3', '4,5,6'])

    def test_singlecolumn(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'csv', 'data': 'A\none\ntwo'},
                'b': {'format': 'csv', 'data': 'A\nthree\nfour'}
            },
            outputs={
                'c': {'format': 'csv'}
            })
        self.assertEqual(outputs['c']['format'], 'csv')
        self.assertEqual(outputs['c']['data'].splitlines(),
                         ['A', 'one', 'two', 'three', 'four'])

    def test_singlerow(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'csv', 'data': 'A,one,two'},
                'b': {'format': 'csv', 'data': 'A,three,four'}
            },
            outputs={
                'c': {'format': 'csv'}
            })
        self.assertEqual(outputs['c']['format'], 'csv')
        # Only the 'header' of the first input is preserved.
        self.assertEqual(outputs['c']['data'].splitlines(),
                         ['A,one,two'])

    def test_tsv(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'csv', 'data': 'a,b,c\n1,2,3'},
                'b': {'format': 'tsv', 'data': 'a\tb\tc\n4\t5\t6'}
            },
            outputs={
                'c': {'format': 'tsv'}
            })
        self.assertEqual(outputs['c']['format'], 'tsv')
        self.assertEqual(outputs['c']['data'].splitlines(),
                         ['a\tb\tc', '1\t2\t3', '4\t5\t6'])

    def test_mongo_to_python(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {
                    'format': 'objectlist.bson',
                    'mode': 'mongodb',
                    'db': 'test',
                    'collection': 'a'
                },
                'b': {
                    'format': 'objectlist.bson',
                    'mode': 'mongodb',
                    'db': 'test',
                    'collection': 'b'
                }
            },
            outputs={
                'c': {'format': 'rows'}
            })
        self.assertEqual(outputs['c']['format'], 'rows')
        self.assertEqual(outputs['c']['data'], {
            'fields': ['_id', 'bar', 'foo'], 'rows': [self.aobj, self.bobj]})

    def test_chaining(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {
                    'format': 'rows',
                    'data': {
                        'fields': ['a', 'b'], 'rows': [{'a': 1, 'b': 2}]}
                    },
                'b': {
                    'format': 'rows',
                    'data': {
                        'fields': ['a', 'b'], 'rows': [{'a': 3, 'b': 4}]
                    }
                }
            },
            outputs={
                'c': {'format': 'rows'}
            })

        outputs = run(
            self.analysis,
            inputs={
                'a': outputs['c'],
                'b': {
                    'format': 'rows',
                    'data': {'fields': ['a', 'b'], 'rows': [{'a': 5, 'b': 6}]}
                }
            },
            outputs={
                'c': {'format': 'rows'}
            })
        self.assertEqual(outputs['c']['format'], 'rows')
        self.assertEqual(outputs['c']['data'], {
            'fields': ['a', 'b'],
            'rows': [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]
        })

    def test_objectlist_to_rows(self):
        objlist = [{'a': {'b': 5}}, {'a': {'b': {'c': 3}}}]
        output = convert('table', {
            'format': 'objectlist',
            'data': objlist
        }, {'format': 'rows'})
        self.assertEqual(output['format'], 'rows')
        self.assertEqual(output['data'], {
            'fields': ['a.b', 'a.b.c'],
            'rows': [{'a.b': 5}, {'a.b.c': 3}]})
        output = convert('table', {
            'format': 'rows',
            'data': output['data']
        }, {'format': 'objectlist'})
        self.assertEqual(output['data'], objlist)

    def test_column_names(self):
        output = convert('table', {
            'format': 'rows',
            'data': {'fields': ['a', 'b'], 'rows': [{'a': 6, 'b': 5}]}
        }, {'format': 'column.names'})
        self.assertEqual(output['format'], 'column.names')
        self.assertEqual(output['data'], ['a', 'b'])

    def test_column_names_csv(self):
        output = convert('table', {
            'format': 'csv',
            'data': ',a,b,longer name\n1,1,1,1\n2,2,2,2\n3,3,3,3\n'
        }, {'format': 'column.names'})
        self.assertEqual(output['format'], 'column.names')
        self.assertEqual(output['data'], ['', 'a', 'b', 'longer name'])

    def test_column_names_discrete(self):
        output = convert('table', {
            'format': 'rows',
            'data': {
                'fields': ['a', 'b', 'disc'],
                'rows': [{'a': 6, 'b': 5, 'disc': 'yes'}]
            }
        }, {'format': 'column.names.discrete'})
        self.assertEqual(output['format'], 'column.names.discrete')
        self.assertEqual(output['data'], ['disc'])

    def test_column_names_continuous(self):
        output = convert('table', {
            'format': 'rows',
            'data': {
                'fields': ['a', 'b', 'disc'],
                'rows': [{'a': 6, 'b': 5, 'disc': 'yes'}]
            }
        }, {'format': 'column.names.continuous'})
        self.assertEqual(output['format'], 'column.names.continuous')
        self.assertEqual(output['data'], ['a', 'b'])

    def test_r_dataframe(self):
        outputs = run(
            self.analysis_r,
            inputs={
                'a': {
                    'format': 'rows',
                    'data': {
                        'fields': ['aa', 'bb'], 'rows': [{'aa': 1, 'bb': 2}]
                    }
                }
            },
            outputs={
                'b': {'format': 'rows'}
            })
        self.assertEqual(outputs['b']['format'], 'rows')
        self.assertEqual(outputs['b']['data'], {
            'fields': ['aa', 'bb'], 'rows': [{'aa': 1, 'bb': 2}]})

    def test_flu(self):
        output = convert(
            'table',
            {
                'format': 'csv',
                'url': 'file://' + os.path.join('data', 'flu.csv')
            },
            {'format': 'column.names'}
        )
        self.assertEqual(output['format'], 'column.names')
        self.assertEqual(len(output['data']), 162)
        self.assertEqual(
            output['data'][:3],
            ['Date', 'United States', 'Alabama']
        )

    def test_header_detection(self):
        output = convert(
            'table',
            {'format': 'csv', 'data': 'a,b,c\n7,1,c\n8,2,f\n9,3,i'},
            {'format': 'rows'}
        )
        self.assertEqual(output['data']['fields'], ['a', 'b', 'c'])
        self.assertEqual(len(output['data']['rows']), 3)

        output = convert(
            'table',
            {'format': 'csv', 'data': '1,2,3\n7,10,\n,11,\n,12,'},
            {'format': 'rows'}
        )
        self.assertEqual(
            output['data']['fields'],
            ['1', '2', '3']
        )
        self.assertEqual(len(output['data']['rows']), 3)

    def test_sniffer(self):
        output = convert(
            'table',
            {
                'format': 'csv',
                'url': 'file://' + os.path.join('data', 'test.csv')
            },
            {'format': 'rows'}
        )
        self.assertEqual(len(output['data']['fields']), 32)
        self.assertEqual(output['data']['fields'][:3], [
            'FACILITY', 'ADDRESS', 'DATE OF INSPECTION'
        ])
        self.assertEqual(len(output['data']['rows']), 14)

        flu = load(os.path.join(
            self.analysis_path, 'xdata', 'flu.json'))

        output = run(
            flu,
            inputs={},
            outputs={'data': {'type': 'table', 'format': 'rows'}}
        )
        self.assertEqual(output['data']['data']['fields'][:3], [
            'Date', 'United States', 'Alabama'
        ])

    def test_big_header(self):
        output = convert(
            'table',
            {
                'format': 'csv',
                'url': 'file://' + os.path.join('data', 'RadiomicsData.csv')
            },
            {'format': 'rows'}
        )
        self.assertEqual(len(output['data']['fields']), 454)
        self.assertEqual(output['data']['fields'][:3], [
            'GLCM_autocorr', 'GLCM_clusProm', 'GLCM_clusShade'
        ])
        self.assertEqual(len(output['data']['rows']), 99)

    def test_nan(self):
        output = convert(
            'table',
            {
                'format': 'csv',
                'url': 'file://' + os.path.join('data', 'RadiomicsData.csv')
            },
            {'format': 'rows.json'}
        )
        data = json.loads(output['data'])
        self.assertEqual(len(data['fields']), 454)
        self.assertEqual(data['fields'][:3], [
            'GLCM_autocorr', 'GLCM_clusProm', 'GLCM_clusShade'
        ])
        self.assertEqual(len(data['rows']), 99)
        for row in data['rows']:
            for field in row:
                if isinstance(row[field], float):
                    self.assertFalse(math.isnan(row[field]))
                    self.assertFalse(math.isinf(row[field]))

    def test_objectlist(self):
        rows = {
            'fields': ['a', 'b'],
            'rows': [
                {'a': 1, 'b': 'x'},
                {'a': 4, 'b': 'y'}
            ]
        }
        objectlist = convert(
            'table',
            {'format': 'rows', 'data': rows},
            {'format': 'objectlist'}
        )['data']
        # Should have same row data
        self.assertEqual(objectlist, rows['rows'])

        rows2 = convert(
            'table',
            {'format': 'objectlist', 'data': objectlist},
            {'format': 'rows'}
        )['data']
        # Should have same fields but could be in different order
        self.assertEqual(set(rows['fields']), set(rows2['fields']))
        # Should have same row data
        self.assertEqual(rows['rows'], rows2['rows'])

        # Make sure we can go back and forth to JSON
        objectlist = convert(
            'table',
            convert(
                'table',
                {'format': 'objectlist', 'data': rows['rows']},
                {'format': 'objectlist.json'}
            ),
            {'format': 'objectlist'}
        )['data']
        self.assertEqual(rows['rows'], objectlist)

    def test_jsonlines(self):
        output = convert('table', {
            'format': 'jsonlines',
            'data': '{"a": 1, "b": 2}\n{"a": 3, "b": 4}'
        }, {'format': 'objectlist'})
        self.assertEqual(output['format'], 'objectlist')
        self.assertEqual(output['data'], [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])


if __name__ == '__main__':
    unittest.main()



import json
import os
from girder_worker.plugins.types import convert
import networkx as nx
from networkx.algorithms.isomorphism import is_isomorphic, numerical_edge_match
import unittest
from lxml import etree


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.GRAPHML_NS = '{http://graphml.graphdrawing.org/xmlns}'
        self.test_input = {
            'distances': {
                'format': 'networkx',
                'data': nx.Graph([
                    ('US', 'UK', {'distance': 4242}),
                    ('US', 'Australia', {'distance': 9429}),
                    ('UK', 'Australia', {'distance': 9443}),
                    ('US', 'Japan', {'distance': 6303})
                ])
            },
            'alphabetGraph': {
                'format': 'clique.json'
            }
        }

        with open(os.path.join('girder_worker', 'core', 'tests',
                               'data', 'clique.json'), 'rb') as fixture:
            self.test_input['alphabetGraph']['data'] = fixture.read()

    def test_clique(self):
        # clique.json -> NetworkX
        output = convert(
            'graph', self.test_input['alphabetGraph'], {'format': 'networkx'})

        self.assertEqual(
            set([n[1]['name'] for n in output['data'].nodes(data=True)]),
            set(['a', 'b', 'c', 'd']))
        self.assertEqual(len(output['data'].edges()), 3)
        self.assertEqual(output['data'].degree('55ba5019f8883b5bf35f3e30'), 0)

        # NetworkX -> clique.json
        output = convert(
            'graph', output, {'format': 'clique.json'})

        # Since the id of the nodes are lost, only test the structure
        # Check nodes with names a, b, c, and d
        # Check the following edges
        # a -> b
        # a -> c
        # b -> c
        output['data'] = json.loads(output['data'])
        nodes = [item for item in output['data'] if item['type'] == 'node']
        edges = [(item['source']['$oid'], item['target']['$oid'])
                 for item in output['data'] if item['type'] == 'link']
        oid_by_name = {}

        for node in nodes:
            oid_by_name[node['data']['name']] = node['_id']['$oid']

        # Check nodes
        self.assertEqual(sorted(oid_by_name.keys()),
                         ['a', 'b', 'c', 'd'])

        # Check edges
        self.assertEqual(len(edges), 3)
        self.assertIn((oid_by_name['a'], oid_by_name['b']), edges)
        self.assertIn((oid_by_name['a'], oid_by_name['c']), edges)
        self.assertIn((oid_by_name['b'], oid_by_name['c']), edges)

    def test_adjacencylist(self):
        output = convert(
            'graph', self.test_input['distances'], {'format': 'adjacencylist'})

        expected_edges = set(self.test_input['distances']['data'].edges())
        actual_edges = set()

        for line in output['data'].splitlines():
            parts = line.split(' ', 1)

            if len(parts) > 1:
                source, targets = parts

                for target in targets.split(' '):
                    edge = (source, target)
                    self.assertNotIn(edge, actual_edges)
                    actual_edges.add(edge)

        self.assertEqual(expected_edges, actual_edges)

        output = convert(
            'graph', output, {'format': 'networkx'})

        # Don't take edges into consideration, because they were lost in the
        # original conversion
        self.assertTrue(
            is_isomorphic(output['data'],
                          self.test_input['distances']['data'],
                          edge_match=None))

    def test_graphml(self):
        output = convert(
            'graph', self.test_input['distances'], {'format': 'graphml'})
        expected_edges = set(self.test_input['distances']['data'].edges(
            data='distance'))
        actual_edges = set()

        self.assertIsInstance(output['data'], (str, unicode))
        tree = etree.fromstring(output['data'])
        self.assertEqual(len(tree), 2)
        self.assertEqual(tree[0].tag, self.GRAPHML_NS + 'key')
        self.assertEqual(tree[1].tag, self.GRAPHML_NS + 'graph')

        for edge in tree[1].findall(self.GRAPHML_NS + 'edge'):
            edge = (edge.attrib['source'],
                    edge.attrib['target'],
                    int(edge.find(self.GRAPHML_NS + 'data').text))

            self.assertNotIn(edge, actual_edges)
            actual_edges.add(edge)

        self.assertEqual(expected_edges, actual_edges)

        output = convert(
            'graph', output, {'format': 'networkx'})

        self.assertTrue(
            is_isomorphic(output['data'],
                          self.test_input['distances']['data'],
                          edge_match=numerical_edge_match('distance', 1)))


if __name__ == '__main__':
    unittest.main()



import events
import io
import json
import os

from executors.python import run as python_run
from executors.workflow import run as workflow_run
from . import utils

from girder_worker.utils import JobStatus, StateTransitionException
from girder_worker import config, PACKAGE_DIR

# Maps task modes to their implementation
_task_map = {}


class TaskSpecValidationError(Exception):
    pass


def register_executor(name, fn):
    """
    Register a new executor in the girder_worker runtime. This is used to
    map the 'mode' field of a task to a function that will execute the task.

    :param name: The value of the mode field that maps to the given function.
    :type name: str
    :param fn: The implementing function.
    :type fn: function
    """
    _task_map[name] = fn


def unregister_executor(name):
    """
    Unregister an executor from the map.

    :param name: The name of the executor to unregister.
    :type name: str
    """
    del _task_map[name]


def _resolve_scripts(task):
    if task.get('mode') != 'workflow':
        if 'script_uri' in task and 'script' not in task:
            task['script'] = io.fetch({
                'url': task['script_uri']
            })
    elif 'steps' in task:
        for step in task['steps']:
            _resolve_scripts(step['task'])


def load(task_file):
    """
    Load a task JSON into memory, resolving any ``'script_uri'`` fields
    by replacing it with a ``'script'`` field containing the contents pointed
    to by ``'script_uri'`` (see :py:mod:`girder_worker.uri` for URI formats). A
    ``script_fetch_mode`` field may also be set

    :param task_file: The path to the JSON file to load.
    :returns: The task as a dictionary.
    """
    with open(task_file) as f:
        task = json.load(f)

    prevdir = os.getcwd()
    parent = os.path.dirname(task_file)
    if parent != '':
        os.chdir(os.path.dirname(task_file))
    _resolve_scripts(task)
    os.chdir(prevdir)

    return task


def set_job_status(mgr, status):
    if mgr:
        mgr.updateStatus(status)


def _extractId(spec):
    return spec['id'] if 'id' in spec else spec['name']


def _validateInputs(task_inputs, inputs):
    for name, task_input in task_inputs.iteritems():
        if name not in inputs:
            if 'default' in task_input:
                inputs[name] = task_input['default']
            else:
                raise Exception('Required input \'%s\' not provided.' % name)


@utils.with_tmpdir
def run(task, inputs=None, outputs=None, fetch=True, status=None, **kwargs):
    """
    Run a task with the specified I/O bindings.

    :param task: Specification of the task to run.
    :type task: dict
    :param inputs: Specification of how input objects should be fetched
        into the runtime environment of this task.
    :type inputs: dict
    :param outputs: Specification of what should be done with outputs
        of this task.
    :type outputs: dict
    :param write_script: If ``True`` task scripts will be written to file before
        being passed to ``exec``. This improves interactive debugging with
        tools such as ``pdb`` at the cost of additional file I/O. Note that
        when passed to run *all* tasks will be written to file including
        validation and conversion tasks.
    :param fetch: If ``True`` will perform a fetch on the input before
        running the task (default ``True``).
    :param status: Job status to update to during execution of this task.
    :type status: girder_worker.utils.JobStatus
    :returns: A dictionary of the form ``name: binding`` where ``name`` is
        the name of the output and ``binding`` is an output binding of the form
        ``{'data': data}``. The ``'data'`` field may be absent if an output URI
        was provided. Instead, those outputs will be saved to that URI and the
        output binding will contain the location in the ``'uri'`` field.
    """
    inputs = inputs or {}
    outputs = outputs or {}

    task_inputs = {_extractId(d): d for d in task.get('inputs', ())}
    task_outputs = {_extractId(d): d for d in task.get('outputs', ())}
    mode = task.get('mode', 'python')

    if mode not in _task_map:
        raise Exception('Invalid mode: %s' % mode)

    job_mgr = kwargs.get('_job_manager')

    info = {
        'task': task,
        'task_inputs': task_inputs,
        'task_outputs': task_outputs,
        'mode': mode,
        'inputs': inputs,
        'outputs': outputs,
        'status': status,
        'job_mgr': job_mgr,
        'kwargs': kwargs
    }
    events.trigger('run.before', info)

    try:
        # If some inputs are not there, fill in with defaults
        _validateInputs(task_inputs, inputs)

        for name, d in inputs.iteritems():
            task_input = task_inputs[name]
            if task_input.get('stream'):
                continue  # this input will be fetched as a stream

            if fetch:
                if status == JobStatus.RUNNING and 'data' not in d:
                    set_job_status(job_mgr, JobStatus.FETCHING_INPUT)
                d['data'] = io.fetch(d, **dict({'task_input': task_input}, **kwargs))

            events.trigger('run.handle_input', {
                'info': info,
                'task_input': task_input,
                'input': d,
                'name': name
            })

            if 'script_data' not in d:
                d['script_data'] = d['data']

        for name, task_output in task_outputs.iteritems():
            if name not in outputs:
                outputs[name] = {}

        # Set the appropriate job status flag
        set_job_status(job_mgr, status)

        # Actually run the task for the given mode
        _task_map[mode](
            task=task, inputs=inputs, outputs=outputs, task_inputs=task_inputs,
            task_outputs=task_outputs, **kwargs)

        for name, task_output in task_outputs.iteritems():
            if task_output.get('stream'):
                continue  # this output has already been sent as a stream

            output = outputs[name]
            e = events.trigger('run.handle_output', {
                'info': info,
                'task_output': task_output,
                'output': output,
                'outputs': outputs,
                'name': name
            })

            if not e.default_prevented:
                data = outputs[name]['script_data']

                if status == JobStatus.RUNNING:
                    set_job_status(job_mgr, JobStatus.PUSHING_OUTPUT)
                io.push(data, outputs[name], **dict({'task_output': task_output}, **kwargs))

            output.pop('script_data', None)

        events.trigger('run.after', info)

        return outputs
    except StateTransitionException:
        if job_mgr:
            status = job_mgr.refreshStatus()
            # If we are canceling we want to stay in that state, otherwise raise
            # the exception
            if status != JobStatus.CANCELING:
                raise
        else:
            raise
    finally:
        events.trigger('run.finally', info)


register_executor('python', python_run)
register_executor('workflow', workflow_run)

# Load plugins that are enabled in the config file or env var
_plugins = os.environ.get('WORKER_PLUGINS_ENABLED',
                          config.get('girder_worker', 'plugins_enabled'))
_plugins = [p.strip() for p in _plugins.split(',') if p.strip()]
_paths = os.environ.get(
    'WORKER_PLUGIN_LOAD_PATH', config.get(
        'girder_worker', 'plugin_load_path')).split(':')
_paths = [p for p in _paths if p.strip()]
_paths.append(os.path.join(PACKAGE_DIR, 'plugins'))
utils.load_plugins(_plugins, _paths, quiet=True)



from girder_worker.tasks import run
import unittest


class TestNumber(unittest.TestCase):

    def setUp(self):
        self.analysis = {
            'name': 'add',
            'inputs': [
                {
                    'name': 'a',
                    'type': 'string',
                    'format': 'text',
                    'default': {
                        'format': 'json',
                        'data': 'hi'
                    }
                },
                {
                    'name': 'b',
                    'type': 'string',
                    'format': 'text'
                }
            ],
            'outputs': [{'name': 'c', 'type': 'string', 'format': 'text'}],
            'script': 'c = a + b',
            'mode': 'python'
        }

    def test_text(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'text', 'data': 'hi, '},
                'b': {'format': 'text', 'data': 'there'}
            },
            outputs={
                'c': {'format': 'text'}
            })
        self.assertEqual(outputs['c']['format'], 'text')
        self.assertEqual(outputs['c']['data'], 'hi, there')

    def test_string(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'string', 'data': 'hi, '},
                'b': {'format': 'text', 'data': 'there'}
            },
            outputs={
                'c': {'format': 'text'}
            })
        self.assertEqual(outputs['c']['format'], 'text')
        self.assertEqual(outputs['c']['data'], 'hi, there')

    def test_json(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'json', 'data': '"hi, "'},
                'b': {'format': 'json', 'data': '"there"'}
            },
            outputs={
                'c': {'format': 'json'}
            })
        self.assertEqual(outputs['c']['format'], 'json')
        self.assertEqual(outputs['c']['data'], '"hi, there"')

    def test_unicode(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'text', 'data': u'hi, '},
                'b': {'format': 'text', 'data': u'there'}
            },
            outputs={
                'c': {'format': 'text'}
            })
        self.assertEqual(outputs['c']['format'], 'text')
        self.assertEqual(outputs['c']['data'], u'hi, there')
        self.assertIsInstance(outputs['c']['data'], unicode)


if __name__ == '__main__':
    unittest.main()



"""Module defining Python testing infrastructure."""

import os
import shutil
import sys
import tempfile
from unittest import TestCase as _TestCase

try:
    import girder_worker as _girder_worker
except Exception:
    sys.path.append(os.path.abspath(os.path.join('..', '..')))
    sys.path.append(os.getcwd())
    import girder_worker as _girder_worker

_modpath = os.path.dirname(os.path.abspath(__file__))


class TestCase(_TestCase):
    """Base testing class extending unittest.TestCase."""

    girder_worker = _girder_worker
    outpath = os.environ.get('CTEST_BINARY_DIRECTORY')
    modpath = _modpath

    def setUp(self):  # noqa
        """Create output directory if necessary."""
        self.deleteoutputdir = False
        if self.outpath is None:
            self.outpath = tempfile.mkdtemp()
            self.deleteoutputdir = True

    def tearDown(self):  # noqa
        """Delete output directory if necessary."""
        if self.deleteoutputdir:
            try:
                shutil.rmtree(self.outputpath)
                self.outputpath = None
                self.deleteoutputdir = False
            except Exception:
                pass

    @staticmethod
    def data_path(filename=None):
        """Return an absolute path to a data file or directory."""
        pth = os.path.join(_modpath, 'data')
        if filename is not None:
            pth = os.path.join(pth, filename)
        return pth

    def output_path(self, filename=None):
        """Return an absolute path to a place to store outputs."""
        pth = self.outpath
        if filename is not None:
            pth = os.path.join(pth, filename)
        return pth



from girder_worker.app import app


@app.task
def fibonacci(n):
    if n == 1 or n == 2:
        return 1
    return fibonacci(n-1) + fibonacci(n-2)


@app.task(bind=True)
def fib_seq(task, n):
    if n < 0:
        raise Exception("Must pass in positive integer!")

    for _n in range(1, n+1):
        print "%s: %s" % (_n, fibonacci(_n))



import contextlib
import six
import sys


@contextlib.contextmanager
def captureOutput():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [six.StringIO(), six.StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()



import imp
import json
import sys
import tempfile


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    custom = imp.new_module('__girder_worker__')

    custom.__dict__['_job_manager'] = kwargs.get('_job_manager')
    custom.__dict__['_tempdir'] = kwargs.get('_tempdir')
    custom.__dict__['_celery_task'] = kwargs.get('_celery_task')

    for name in inputs:
        custom.__dict__[name] = inputs[name]['script_data']

    if task.get('write_script', kwargs.get('write_script', False)):
        debug_path = tempfile.mktemp()
        with open(debug_path, 'wb') as fh:
            fh.write(task['script'])

        with open(debug_path, 'r') as fh:
            exec fh in custom.__dict__

    else:
        try:
            exec task['script'] in custom.__dict__
        except Exception, e:
            trace = sys.exc_info()[2]
            lines = task['script'].split('\n')
            lines = [(str(i+1) + ': ' + lines[i]) for i in xrange(len(lines))]
            error = (
                str(e) + '\nScript:\n' + '\n'.join(lines) +
                '\nTask:\n' + json.dumps(task, indent=4)
            )
            raise Exception(error), None, trace

    for name, task_output in task_outputs.iteritems():
        outputs[name]['script_data'] = custom.__dict__[name]



import base64
from girder_worker.tasks import run
from girder_worker.plugins.types import convert
import math
import operator
import os
import tempfile
import unittest
from PIL import Image
from PIL.JpegImagePlugin import JpegImageFile
from six import StringIO


def compareImages(im1, im2):
    h1 = im1.histogram()
    h2 = im2.histogram()

    return math.sqrt(reduce(operator.add,
                     map(lambda a, b: (a-b)**2, h1, h2))/len(h1))


class TestImage(unittest.TestCase):

    def setUp(self):
        self.image = (
            'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAAK/INwWK6'
            'QAAABl0RVh0U29mdHdhcmUAQWRvYmUgSW1hZ2VSZWFkeXHJZTwAAAKfSURBVDjL'
            'pZPrS1NhHMf9O3bOdmwDCWREIYKEUHsVJBI7mg3FvCxL09290jZj2EyLMnJexkg'
            'pLbPUanNOberU5taUMnHZUULMvelCtWF0sW/n7MVMEiN64AsPD8/n83uucQDi/i'
            'd/DBT4Dolypw/qsz0pTMbj/WHpiDgsdSUyUmeiPt2+V7SrIM+bSss8ySGdR4abQ'
            'Qv6lrui6VxsRonrGCS9VEjSQ9E7CtiqdOZ4UuTqnBHO1X7YXl6Daa4yGq7vWO1D'
            '40wVDtj4kWQbn94myPGkCDPdSesczE2sCZShwl8CzcwZ6NiUs6n2nYX99T1cnKq'
            'A2EKui6+TwphA5k4yqMayopU5mANV3lNQTBdCMVUA9VQh3GuDMHiVcLCS3J4jSL'
            'hCGmKCjBEx0xlshjXYhApfMZRP5CyYD+UkG08+xt+4wLVQZA1tzxthm2tEfD3Jx'
            'ARH7QkbD1ZuozaggdZbxK5kAIsf5qGaKMTY2lAU/rH5HW3PLsEwUYy+YCcERmIj'
            'JpDcpzb6l7th9KtQ69fi09ePUej9l7cx2DJbD7UrG3r3afQHOyCo+V3QQzE35pv'
            'QvnAZukk5zL5qRL59jsKbPzdheXoBZc4saFhBS6AO7V4zqCpiawuptwQG+UAa7C'
            't3UT0hh9p9EnXT5Vh6t4C22QaUDh6HwnECOmcO7K+6kW49DKqS2DrEZCtfuI+9G'
            'rNHg4fMHVSO5kE7nAPVkAxKBxcOzsajpS4Yh4ohUPPWKTUh3PaQEptIOr6BiJjc'
            'ZXCwktaAGfrRIpwblqOV3YKdhfXOIvBLeREWpnd8ynsaSJoyESFphwTtfjN6X1j'
            'RO2+FxWtCWksqBApeiFIR9K6fiTpPiigDoadqCEag5YUFKl6Yrciw0VOlhOivv/'
            'Ff8wtn0KzlebrUYwAAAABJRU5ErkJggg==')

        self.analysis = {
            'name': 'Image pixels',
            'inputs': [
                {
                    'name': 'a',
                    'type': 'image',
                    'format': 'pil'
                }
            ],
            'outputs': [
                {
                    'name': 'pixels',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'script': 'pixels = a.size[0] * a.size[1]\n',
            'mode': 'python'
        }

    def test_base64(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'png.base64', 'data': self.image},
            })
        self.assertEqual(outputs['pixels']['format'], 'number')
        self.assertEqual(outputs['pixels']['data'], 256)

    def test_convert(self):
        tmp = tempfile.mktemp()

        output = convert('image', {
            'format': 'png.base64',
            'data': self.image
        }, {
            'format': 'png',
            'url': 'file://' + tmp,
            'mode': 'auto'
        })

        value = open(tmp).read()
        os.remove(tmp)
        self.assertEqual(output['format'], 'png')
        self.assertEqual(base64.b64encode(value), self.image)

        output = convert(
            'image',
            {'format': 'png.base64', 'data': self.image},
            {'format': 'pil'})

        tmp = tempfile.mktemp()

        output = convert(
            'image',
            output,
            {'format': 'png'})

        io1 = StringIO(base64.b64decode(self.image))
        im1 = Image.open(io1)
        io2 = StringIO(output['data'])
        im2 = Image.open(io2)
        self.assertEqual(compareImages(im1, im2), 0)

        output = convert('image', {
            'format': 'png.base64',
            'data': self.image
        }, {
            'format': 'jpeg'
        })
        data = StringIO(output['data'])
        jpeg = Image.open(data)
        self.assertTrue(isinstance(jpeg, JpegImageFile))


if __name__ == '__main__':
    unittest.main()



import sys
import importlib
import doctest

from base import TestCase  # noqa


if __name__ == '__main__':

    error = 0
    total = 0
    fail = 0
    for module in sys.argv[1:]:
        if module == '-v':
            continue
        try:
            mod = importlib.import_module(module)
        except ImportError:
            error += 1
            print('Could not import {0}.'.format(module))
            continue
        result = doctest.testmod(mod)
        fail += result[0]
        total += result[1]

    if error:
        print('{0} modules failed to import'.format(error))
    print('{0} tests failed out of {1}'.format(fail, total))
    if error or fail:
        sys.exit(1)



from importlib import import_module
import celery
from girder_worker_utils import decorators

import six

from stevedore import extension


#: Defines the namespace used for plugin entrypoints
NAMESPACE = 'girder_worker_plugins'

# Define an internal registry for extensions that aren't associated
# with an entrypoint.  This registry is primarily useful for testing
# where one doesn't necessarily want to install a package into the
# environment for a one off test task.
_extensions = {}


def _import_module(module_name):
    """Try to import a module given as a string."""
    module = None
    try:
        module = import_module(module_name)
    except ImportError:
        import traceback
        traceback.print_exc()
        print('Problem importing %s' % module_name)
    return module


def _handle_entrypoint_errors(mgr, entrypoint, exc):
    print('Problem loading plugin %s, skipping' % entrypoint.name)


def get_extension_manager(app=None):
    """Get an extension manager for the plugin namespace."""
    if app is None:
        app = celery.current_app

    return extension.ExtensionManager(
        namespace=NAMESPACE,
        invoke_on_load=True,
        invoke_args=(app,),
        on_load_failure_callback=_handle_entrypoint_errors
    )


def get_task_imports(ext):
    """Return a list of task modules provided by an extension."""
    try:
        includes = ext.obj.task_imports()
    except Exception:
        import traceback
        traceback.print_exc()
        print('Problem instantiating plugin %s, skipping' % ext.name)
        includes = []
    return includes


# Core tasks are only supported on python 2.  When running in python 3
# this always returns an empty list to avoid loading `girder_worker.core`.
def get_core_task_modules(app=None):
    """Return task modules defined by core."""
    if six.PY2:
        return get_task_imports(get_extension_manager(app=app)['core'])
    return []


def get_plugin_task_modules(app=None):
    """Return task modules defined by plugins."""
    includes = []
    for ext in get_extension_manager(app=app):
        if ext.name != 'core':
            includes.extend(get_task_imports(ext))
    return includes


def import_all_includes(core=True):
    """Import all task modules for their side-effects."""
    if core:
        for module in get_core_task_modules():
            _import_module(module)

    for module in get_plugin_task_modules():
        _import_module(module)


def get_extensions(app=None):
    """Get a list of installed extensions."""
    extensions = [ext.name for ext in get_extension_manager(app)] + list(_extensions.keys())

    # Because the "core" entrypoint is installed with girder_worker, we have
    # to manually exclude it from the list of extensions when not on python 2.
    if not six.PY2:
        extensions = list(filter(lambda extension: extension != 'core', extensions))
    return extensions


def get_module_tasks(module_name):
    """Get all tasks defined in a python module.

    :param str module_name: The importable module name
    """
    module = _import_module(module_name)
    tasks = {}

    if module is None:
        return tasks

    for name, func in six.iteritems(vars(module)):
        full_name = '%s.%s' % (module_name, name)
        if not hasattr(func, '__call__'):
            # filter out objects that are not callable
            continue

        try:
            decorators.get_description_attribute(func)
            tasks[full_name] = func
        except decorators.MissingDescriptionException:
            pass
    return tasks


def get_extension_tasks(extension, app=None, celery_only=False):
    """Get the tasks defined by a girder_worker extension.

    :param str extension: The extension name
    :param app: The celery app instance
    :param bool celery_only: If true, only return celery tasks
    """

    tasks = {}
    if extension in _extensions:
        tasks = _extensions[extension]
    else:
        manager = get_extension_manager(app)
        imports = get_task_imports(manager[extension])
        tasks = {}
        for module_name in imports:
            tasks.update(get_module_tasks(module_name))

    if celery_only:  # filter celery tasks
        if app is None:
            from .app import app
        tasks = {
            key: tasks[key] for key in tasks if key in app.tasks
        }

    return tasks


def register_extension(name, tasks):
    """Register an extension by name.

    This is an alternative to registering extensions via
    entrypoints.  Its primary use case is for one off tasks
    used in testing.  Using entrypoint extensions is preferred
    because they will be automatically registered, and the
    tasks will have fully resolved python modules that can
    be reused by other tasks and extensions.

    :param str name: The extension name
    :param dict tasks: A mapping from task name to the task function
    """
    global _extensions
    _extensions[name] = tasks


def discover_tasks(app, core=True):
    if core:
        app.conf.update({
            'CELERY_IMPORTS': get_core_task_modules()
        })

    app.conf.update({
        'CELERY_INCLUDE': get_plugin_task_modules()
    })



import sys
import unittest
from girder_worker.tasks import run
from girder_worker.plugins.types.format import (
    conv_graph, converter_path, has_converter, Validator, print_conversion_graph,
    print_conversion_table)
from six import StringIO
from networkx import NetworkXNoPath


class TestFormat(unittest.TestCase):
    def setUp(self):
        self.stringTextValidator = Validator('string', 'text')

        self.prev_stdout = sys.stdout
        sys.stdout = StringIO()

    def tearDown(self):
        sys.stdout = self.prev_stdout

    def test_converter_path(self):
        with self.assertRaisesRegexp(Exception, 'No such validator foo/bar'):
            converter_path(Validator('foo', 'bar'), Validator('foo', 'baz'))

        # There is no path for types which lie in different components
        with self.assertRaises(NetworkXNoPath):
            converter_path(self.stringTextValidator,
                           Validator('graph', 'networkx'))

        self.assertEquals(converter_path(self.stringTextValidator,
                                         self.stringTextValidator), [])

        # There is a two-step path converting between these types
        self.assertEquals(len(converter_path(self.stringTextValidator,
                                             Validator('string', 'json'))), 2)

    def test_run_exceptions(self):
        number_copy = {
            'inputs': [
                {'name': 'some_number',
                 'type': 'number',
                 'format': 'number'}
            ],
            'outputs': [
                {'name': 'some_number',
                 'type': 'number',
                 'format': 'number'}
            ],
            'script': 'some_number = some_number'
        }

        # If an input has an invalid format, it should throw an exception
        # with the name
        number_copy['inputs'][0]['format'] = 'invalidformat'
        with self.assertRaisesRegexp(
                Exception,
                'some_number: No such validator number/invalidformat'):
            run(number_copy, inputs={'some_number':
                                     {'format': 'json',
                                      'data': '5'}})

        # If the input is valid, but there is no conversion path
        number_copy['inputs'][0]['format'] = 'validformat'
        conv_graph.add_node(Validator('number', 'validformat'))
        with self.assertRaisesRegexp(
                Exception,
                ('some_number: No conversion path from number/json '
                 'to number/validformat')):
            run(number_copy, inputs={'some_number':
                                     {'format': 'json',
                                      'data': '5'}})
        conv_graph.remove_node(Validator('number', 'validformat'))

    def test_is_valid(self):
        self.assertEquals(Validator('string', None).is_valid(), True)

        self.assertEquals(Validator('string', 'json').is_valid(), True)

        self.assertEquals(Validator('invalid_type', None).is_valid(), False)

        self.assertEquals(Validator('invalid_type',
                                    'invalid_format').is_valid(), False)

        self.assertEquals(Validator('string',
                                    'invalid_format').is_valid(), False)

        self.assertEquals(Validator('invalid_type',
                                    'json').is_valid(), False)

    def test_has_converter(self):
        # There are converters from string type
        self.assertTrue(has_converter(Validator('string', None)))

        # There are no converters from string/non-existent-format
        self.assertFalse(has_converter(Validator(
            'string', 'non-existent-format')))

        # There are converters from string/text
        self.assertTrue(has_converter(self.stringTextValidator))

        # There are no converters from string/text to number/anything
        self.assertFalse(has_converter(self.stringTextValidator,
                                       Validator('number', format=None)))

        # There is a converter from string/text -> string/json
        self.assertTrue(has_converter(self.stringTextValidator,
                                      Validator('string', 'json')))

        # There are no self loops in the conversion graph
        self.assertFalse(has_converter(self.stringTextValidator,
                                       self.stringTextValidator))

        # Converters don't go from one type to another
        self.assertFalse(has_converter(Validator('string', format=None),
                                       Validator('number', format=None)))

        # This only has a converter which is multiple hops away
        self.assertTrue(has_converter(
            Validator('table', format='rows'),
            Validator('table', format='objectlist.json')))

    def test_conversion_graph(self):
        print_conversion_graph()

    def test_conversion_table(self):
        print_conversion_table()



from girder_worker.tasks import run
import os
import shutil
import tarfile
import unittest
import zipfile

DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')


class TestDirectory(unittest.TestCase):
    """
    Tests the conversions for the directory type formats.
    """
    extractTask = {
        'inputs': [{
            'id': 'myDir',
            'type': 'directory',
            'format': 'path'
        }],
        'outputs': [{
            'id': 'myDir',
            'type': 'string',
            'format': 'text'
        }, {
            'name': 'files',
            'type': 'python',
            'format': 'object'
        }],
        'script': '\n'.join([
            'import os',
            'files = list(os.listdir(myDir))'
        ]),
        'mode': 'python'
    }

    zipFile = os.path.join(DATA_DIR, 'testZip.zip')
    zipDir = os.path.join(DATA_DIR, 'testZip')
    tarFile = os.path.join(DATA_DIR, 'testTar.tar.gz')
    tarDir = os.path.join(DATA_DIR, 'testTar.tar')
    createFromDir = os.path.join(DATA_DIR, 'shapefile')

    def setUp(self):
        self._cleanArchiveData()

    def tearDown(self):
        self._cleanArchiveData()

    def _cleanArchiveData(self):
        if os.path.exists(self.zipDir):
            shutil.rmtree(self.zipDir)
        if os.path.exists(self.tarDir):
            shutil.rmtree(self.tarDir)
        if os.path.exists(self.createFromDir + '.zip'):
            os.remove(self.createFromDir + '.zip')
        if os.path.exists(self.createFromDir + '.tgz'):
            os.remove(self.createFromDir + '.tgz')

    def testExtractZip(self):
        outputs = run(
            self.extractTask,
            inputs={
                'myDir': {
                    'format': 'zip',
                    'data': self.zipFile
                }
            })

        self.assertEqual(set(outputs['files']['data']),
                         set(('provision.retry', 'site.retry')))
        self.assertEqual(outputs['myDir']['data'], self.zipDir)
        self.assertTrue(os.path.isdir(self.zipDir))
        self.assertTrue(
            os.path.isfile(os.path.join(self.zipDir, 'site.retry')))
        self.assertTrue(
            os.path.isfile(os.path.join(self.zipDir, 'provision.retry')))

    def testExtractTgz(self):
        outputs = run(
            self.extractTask,
            inputs={
                'myDir': {
                    'format': 'tgz',
                    'data': self.tarFile
                }
            })

        self.assertEqual(set(outputs['files']['data']),
                         set(('provision.retry', 'site.retry')))
        self.assertEqual(outputs['myDir']['data'], self.tarDir)
        self.assertTrue(os.path.isdir(self.tarDir))
        self.assertTrue(
            os.path.isfile(os.path.join(self.tarDir, 'site.retry')))
        self.assertTrue(
            os.path.isfile(os.path.join(self.tarDir, 'provision.retry')))

    def testCreateZip(self):
        task = {
            'inputs': [{
                'id': 'myArchive',
                'type': 'directory',
                'format': 'zip'
            }],
            'outputs': [{
                'id': 'output',
                'type': 'string',
                'format': 'text'
            }],
            'script': 'output = myArchive',
            'mode': 'python'
        }

        outputs = run(task, inputs={
            'myArchive': {
                'format': 'path',
                'data': self.createFromDir
            }
        })

        path = outputs['output']['data']
        self.assertEqual(path, os.path.join(DATA_DIR, 'shapefile.zip'))
        self.assertTrue(os.path.isfile(path))
        with zipfile.ZipFile(path, 'r') as zf:
            names = zf.namelist()
            self.assertTrue('shapefile/shapefile.cpg' in names)
            self.assertTrue('shapefile/shapefile.prj' in names)

    def testCreateTgz(self):
        task = {
            'inputs': [{
                'id': 'myArchive',
                'type': 'directory',
                'format': 'tgz'
            }],
            'outputs': [{
                'id': 'output',
                'type': 'string',
                'format': 'text'
            }],
            'script': 'output = myArchive',
            'mode': 'python'
        }

        outputs = run(task, inputs={
            'myArchive': {
                'format': 'path',
                'data': self.createFromDir
            }
        })

        path = outputs['output']['data']
        self.assertEqual(path, os.path.join(DATA_DIR, 'shapefile.tgz'))
        self.assertTrue(os.path.isfile(path))
        with tarfile.open(path, 'r') as tf:
            names = tf.getnames()
            self.assertTrue('shapefile/shapefile.cpg' in names)
            self.assertTrue('shapefile/shapefile.prj' in names)


if __name__ == '__main__':
    unittest.main()



import json
from celery import current_app
from girder_client import GirderClient
from girder_worker import logger
from girder_worker.utils import _maybe_model_repr
from girder_worker_utils import _walk_obj

import requests


class MissingJobArguments(RuntimeError):
    pass


def create_task_job(job_defaults, sender=None, body=None,
                    exchange=None, routing_key=None, headers=None,
                    properties=None, declare=None, retry_policy=None,
                    **kwargs):
    parent_task = current_app.current_task
    try:
        if parent_task is None:
            raise MissingJobArguments('Parent task is None')
        if parent_task.request is None:
            raise MissingJobArguments("Parent task's request is None")
        if not hasattr(parent_task.request, 'girder_api_url'):
            raise MissingJobArguments(
                "Parent task's request does not contain girder_api_url")
        if not hasattr(parent_task.request, 'girder_client_token'):
            raise MissingJobArguments(
                "Parent task's request does not contain girder_client_token")
        if not hasattr(parent_task.request, 'id'):
            raise MissingJobArguments(
                "Parent task's request does not contain id")
        if 'id' not in headers:
            raise MissingJobArguments('id is not in headers')

        gc = GirderClient(apiUrl=parent_task.request.girder_api_url)
        gc.token = parent_task.request.girder_client_token

        task_args = tuple(_walk_obj(body[0], _maybe_model_repr))
        task_kwargs = _walk_obj(body[1], _maybe_model_repr)
        parameters = {
            'title': headers.pop('girder_job_title',
                                 job_defaults.get('girder_job_title', '')),
            'type': headers.pop('girder_job_type',
                                job_defaults.get('girder_job_type', '')),
            'handler': headers.pop('girder_job_handler',
                                   job_defaults.get('girder_job_handler', '')),
            'public': headers.pop('girder_job_public',
                                  job_defaults.get('girder_job_public', '')),
            'args': json.dumps(task_args),
            'kwargs': task_kwargs,
            'otherFields': json.dumps(
                dict(celeryTaskId=headers['id'],
                     celeryParentTaskId=parent_task.request.id,
                     **headers.pop('girder_job_other_fields',
                                   job_defaults.get('girder_job_other_fields', ''))))
        }

        try:
            response = gc.post('job', parameters=parameters, jsonResp=False)
            if response.ok:
                headers['jobInfoSpec'] = response.json().get('jobInfoSpec')
        except requests.exceptions.RequestException as e:
            logger.warn('Failed to post job: {}'.format(e))

    except MissingJobArguments as e:
        logger.warn('Girder job not created: {}'.format(str(e)))


def attach_girder_api_url(sender=None, body=None, exchange=None,
                          routing_key=None, headers=None, properties=None,
                          declare=None, retry_policy=None, **kwargs):
    parent_task = current_app.current_task
    try:
        if parent_task is None:
            raise MissingJobArguments('Parent task is None')
        if parent_task.request is None:
            raise MissingJobArguments("Parent task's request is None")
        if not hasattr(parent_task.request, 'girder_api_url'):
            raise MissingJobArguments(
                "Parent task's request does not contain girder_api_url")
        headers['girder_api_url'] = parent_task.request.girder_api_url
    except MissingJobArguments as e:
        logger.warn('Could not get girder_api_url from parent task: {}'.format(str(e)))


def attach_girder_client_token(sender=None, body=None, exchange=None,
                               routing_key=None, headers=None, properties=None,
                               declare=None, retry_policy=None, **kwargs):
    parent_task = current_app.current_task
    try:
        if parent_task is None:
            raise MissingJobArguments('Parent task is None')
        if parent_task.request is None:
            raise MissingJobArguments("Parent task's request is None")
        if not hasattr(parent_task.request, 'girder_client_token'):
            raise MissingJobArguments(
                "Parent task's request does not contain girder_client_token")

        headers['girder_client_token'] = parent_task.request.girder_client_token
    except MissingJobArguments as e:
        logger.warn('Could not get token from parent task: {}'.format(str(e)))


def get_async_result_job_property(async_result):
    # NOT IMPLEMENTED!
    return None



from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, filtermodel

class Fib(Resource):
    def __init__(self):
        super(Fib, self).__init__()
        self.resourceName = 'fib'

        self.route('POST', (), self.run_fib_sequence)
        self.route('POST', ('gwrun', ), self.run_gw_task)
        self.route('POST', ('gwdocker', ), self.run_docker_task)

    @access.token
    @filtermodel(model='job', plugin='jobs')
    @describeRoute(
        Description('Fire off a a fibonacci sequence job')
        .param('number', 'The upper bound of the fibonacci sequence',
               dataType='integer')
    )
    def run_fib_sequence(self, params):
        number = int(params.get('number', 10))
        jobModel = self.model('job', 'jobs')
        job = jobModel.createJob(
            title='Fib_Seq', type='worker', handler='worker_handler',
            user=self.getCurrentUser(), public=False, args=(number,), kwargs={},
            otherFields={
                'celeryTaskName': 'gwexample.analyses.tasks.fib_seq'
            })

        jobModel.save(job)
        jobModel.scheduleJob(job)

        return job

    @access.token
    @filtermodel(model='job', plugin='jobs')
    @describeRoute(
        Description('Fire off a girder_worker.run job')
    )
    def run_gw_task(self, params):
        analysis = {
            'name': 'add',
            'inputs': [
                {
                    'name': 'a',
                    'type': 'integer',
                    'format': 'integer',
                    'default': {
                        'format': 'json',
                        'data': '0'
                    }
                },
                {
                    'name': 'b',
                    'type': 'integer',
                    'format': 'integer'
                }
            ],
            'outputs': [{'name': 'c', 'type': 'integer', 'format': 'integer'}],
            'script': 'c = a + b',
            'mode': 'python'}

        inputs={'a': {'format': 'integer', 'data': 1},
                'b': {'format': 'integer', 'data': 2}}

        outputs={'c': {'format': 'integer'}}


        jobModel = self.model('job', 'jobs')
        job = jobModel.createJob(
            title='GW_run', type='worker', handler='worker_handler',
            user=self.getCurrentUser(), public=False, args=(analysis,),
            kwargs={'inputs': inputs, 'outputs': outputs})

        jobModel.save(job)
        jobModel.scheduleJob(job)

        return job

    @access.token
    @filtermodel(model='job', plugin='jobs')
    @describeRoute(
        Description('Fire off a docker job')
    )
    def run_docker_task(self, params):
        test_image = 'girder/girder_worker_test:latest'
        task = {
            'mode': 'docker',
            'docker_image': test_image,
            'pull_image': True,
            'container_args': ['$input{test_mode}', '$input{message}'],
            'inputs': [{
                'id': 'test_mode',
                'name': '',
                'format': 'string',
                'type': 'string'
            }, {
                'id': 'message',
                'name': '',
                'format': 'string',
                'type': 'string'
            }],
            'outputs': []
        }

        inputs = {
            'test_mode': {
                'format': 'string',
                'data': 'stdio'
            },
            'message': {
                'format': 'string',
                'data': "Hello from girder_worker docker test!"
            }
        }
        outputs = []

        jobModel = self.model('job', 'jobs')
        job = jobModel.createJob(
            title='GW_docker', type='worker', handler='worker_handler',
            user=self.getCurrentUser(), public=False, args=(task,),
            kwargs={'inputs': inputs, 'outputs': outputs})

        jobModel.save(job)
        jobModel.scheduleJob(job)

        return job


def load(info):
    info['apiRoot'].fib = Fib()



import bson
import os
from girder_worker.tasks import run
from girder_worker.plugins.types import convert
import unittest


class TestTree(unittest.TestCase):

    def setUp(self):
        self.analysis = {
            'name': 'tree_copy',
            'inputs': [{'name': 'a', 'type': 'tree', 'format': 'nested'}],
            'outputs': [{'name': 'b', 'type': 'tree', 'format': 'nested'}],
            'script': 'b = a'
        }
        self.analysis_r = {
            'name': 'tree_copy_r',
            'inputs': [{'name': 'a', 'type': 'tree', 'format': 'r.apetree'}],
            'outputs': [{'name': 'b', 'type': 'tree', 'format': 'r.apetree'}],
            'script': 'b <- a',
            'mode': 'r'
        }
        self.newick = '((ahli:0,allogus:1):2,rubribarbus:3);'
        self.nexus = """#NEXUS


BEGIN TAXA;
    DIMENSIONS NTAX = 4;
    TAXLABELS
        A
        B
        C
        D
    ;
END;
BEGIN TREES;
    TRANSLATE
        1   A,
        2   B,
        3   C,
        4   D
    ;
    TREE * UNTITLED = [&R] ((1:1,2:1):1,(3:1,4:1):1);
END;"""
        # Change dir for loading analyses
        self.prevdir = os.getcwd()
        cur_path = os.path.dirname(os.path.realpath(__file__))
        os.chdir(cur_path)

    def tearDown(self):
        os.chdir(self.prevdir)

    def test_newick(self):
        outputs = run(
            self.analysis,
            inputs={'a': {'format': 'newick', 'data': self.newick}},
            outputs={'b': {'format': 'newick'}}
        )
        self.assertEqual(outputs['b']['format'], 'newick')
        self.assertEqual(outputs['b']['data'], self.newick)

    def test_json(self):
        outputs = run(
            self.analysis,
            inputs={'a': {'format': 'newick', 'data': self.newick}},
            outputs={'b': {'format': 'nested'}}
        )
        self.assertEqual(outputs['b']['format'], 'nested')

        expected = {
            'edge_fields': ['weight'],
            'node_fields': ['node name', 'node weight'],
            'node_data': {'node name': '', 'node weight': 0.0},
            'children': [
                {
                    'node_data': {'node name': '', 'node weight': 2.0},
                    'edge_data': {'weight': 2.0},
                    'children': [
                        {
                            'node_data': {
                                'node name': 'ahli',
                                'node weight': 2.0
                            },
                            'edge_data': {'weight': 0.0}
                        },
                        {
                            'node_data': {
                                'node name': 'allogus',
                                'node weight': 3.0
                            },
                            'edge_data': {'weight': 1.0}
                        }
                    ]
                },
                {
                    'node_data': {
                        'node name': 'rubribarbus',
                        'node weight': 3.0
                    },
                    'edge_data': {'weight': 3.0}
                }
            ]
        }

        self.assertEqual(
            outputs['b']['data'], expected)

    def test_r_apetree(self):
        outputs = run(
            self.analysis,
            inputs={'a': {'format': 'newick', 'data': self.newick}},
            outputs={'b': {'format': 'r.apetree'}}
        )
        self.assertEqual(outputs['b']['format'], 'r.apetree')
        self.assertEqual(
            str(outputs['b']['data'])[:52],
            '\nPhylogenetic tree with 3 tips and 2 internal nodes.')

    def test_r(self):
        outputs = run(
            self.analysis_r,
            inputs={'a': {'format': 'newick', 'data': self.newick}},
            outputs={'b': {'format': 'newick'}}
        )
        self.assertEqual(outputs['b']['format'], 'newick')
        self.assertEqual(outputs['b']['data'], self.newick)

        outputs = run(
            self.analysis_r,
            inputs={'a': {'format': 'nexus', 'data': self.nexus}},
            outputs={'b': {'format': 'nexus'}}
        )
        self.assertEqual(outputs['b']['format'], 'nexus')

        # Ignore spaces vs. tabs, and skip timestamp comment on line 2
        out = '\n'.join(outputs['b']['data'].splitlines()[2:])
        out = ' '.join(out.split())
        expected = '\n'.join(self.nexus.splitlines()[2:])
        expected = ' '.join(expected.split())
        self.assertEqual(out, expected)

    def test_non_binary_tree(self):
        convert(
            'tree',
            {
                'format': 'newick',
                'url': 'file://' +
                       os.path.join('data', 'geospiza_from_otl.phy')
            },
            {'format': 'nested'})

    def test_treestore(self):
        output = convert(
            'tree',
            {'format': 'newick', 'data': self.newick},
            {'format': 'r.apetree'})
        output = convert('tree', output, {'format': 'treestore'})
        self.assertEqual(output['format'], 'treestore')
        rows = bson.decode_all(output['data'])
        for d in rows:
            if 'rooted' in d:
                root = d
        self.assertNotEqual(root, None)
        self.assertEqual(len(root['clades']), 1)

        def findId(id):
            for d in rows:
                if d['_id'] == id:
                    return d

        top = findId(root['clades'][0])
        self.assertEqual(len(top['clades']), 2)
        internal = findId(top['clades'][0])
        rubribarbus = findId(top['clades'][1])
        ahli = findId(internal['clades'][0])
        allogus = findId(internal['clades'][1])
        self.assertEqual(internal['branch_length'], 2)
        self.assertEqual(ahli['name'], 'ahli')
        self.assertEqual(ahli['branch_length'], 0)
        self.assertEqual(allogus['name'], 'allogus')
        self.assertEqual(allogus['branch_length'], 1)
        self.assertEqual(rubribarbus['name'], 'rubribarbus')
        self.assertEqual(rubribarbus['branch_length'], 3)



"""This module defines I/O ports that serve as interfaces between tasks."""

import six

from girder_worker.core import io
from girder_worker.plugins.types import convert, isvalid, format
from .spec import Spec


class ValidationError(Exception):
    """An exception type raised when encountering invalid data types."""

    message_format = (
        'Input "{name}" (Python type "{python_type}") is not of the '
        'expected type ("{type}") and format ("{format}")'
    )

    def __init__(self, port, data_spec):
        """Generate a data validation exception.

        :param port: The port that encountered the error
        :type port: :py:class:Port
        :param dict data_spec: The data specification passed to the port.
        """
        self.port = port
        self.data_spec = data_spec

    def __str__(self):
        """Initialize an error message for the exception."""
        return self.message_format.format(
            name=str(self.port.name),
            python_type=str(type(self.data_spec.get('data'))),
            type=str(self.port.type),
            format=str(self.port.format)
        )


class Port(Spec):
    """A port defines a communication channel between tasks.

    Ports enable bidirectional communication between tasks and are responsible
    for ensuring that the connections are compatible.  The primary purpose of
    ports is to specify what types of data tasks can read and write.  This
    information is used by tasks to determine if they can be connected.  Ports
    also provide documentation for the task by describing its inputs and
    outputs. Ports also handle fetching data from and pushing data to remote
    data stores.

    >>> spec = {'name': 'a', 'type': 'number', 'format': 'number'}
    >>> port = Port(spec)

    The port object is serialized as a json object

    >>> import json
    >>> json.loads(str(port)) == spec
    True

    It has several properties derived from the spec

    >>> port.name == spec['name']
    True
    >>> port.type == spec['type']
    True
    >>> port.format == spec['format']
    True

    It also supports auto converting formats and validation by default

    >>> port.auto_convert
    True
    >>> port.auto_validate
    True

    Spec properties are automatically validated when setting them

    >>> port = Port()
    Traceback (most recent call last):
        ...
    ValueError: Port specs require a valid name.

    >>> port = Port(name="my port", type="python", format="object")
    >>> port.format = 'invalid'
    Traceback (most recent call last):
        ...
    ValueError: Unknown format "python.invalid"

    Checking the ``type`` is deferred to allow incremental updating

    >>> port['type'] = 'image'
    >>> port.json()
    Traceback (most recent call last):
        ...
    ValueError: Unknown format "image.object"

    >>> port.format = 'png'
    >>> port.json()
    '{"type": "image", "name": "my port", "format": "png"}'
    >>> port == Port(port)
    True
    """

    def __init__(self, *arg, **kw):
        """Initialize the port on a given task.

        Extends the spec initialization by appending defaults and adding basic
        validation.  By default, port specs take "python.object" data.
        """
        super(Port, self).__init__(*arg, **kw)
        self.add_validation_check('Port.name', Port.__check_name)
        self.add_validation_check('Port.type', Port.__check_types)
        self.check()

    def __check_name(self, key=None, oldvalue=None, newvalue=None, **kw):
        """Ensure that the spec has necessary keys."""
        if 'name' not in self or not isinstance(self['name'], six.string_types):
            raise ValueError('Port specs require a valid name.')

    def __check_types(self, key=None, oldvalue=None, newvalue=None, **kw):
        """Ensure the data format given is known."""
        if key in ('type', None) and not format.Validator(
                self['type'], None).is_valid():
            raise ValueError('Unknown type "%s"' % (self['type'],))

        elif key in ('format', None) and not format.Validator(
                self['type'], self['format']).is_valid():
            raise ValueError(
                'Unknown format "%s.%s"' % (self['type'], self['format'])
            )

    def validate(self, data_spec):
        """Ensure the given data spec is compatible with this port.

        :param dict data_spec: Data specification
        :returns: bool

        >>> spec = {'name': 'a', 'type': 'number', 'format': 'number'}
        >>> port = Port(spec)

        >>> port.validate({'format': 'number', 'data': 1.5})
        True
        >>> port.validate({'format': 'json', 'data': '1.5'})
        True
        >>> port.validate({'format': 'number', 'data': '1.5'})
        False
        >>> port.validate({'format': 'unknown format', 'data': '...'})
        False
        """
        try:
            return isvalid(self.type, data_spec)
        except Exception:  # catchall validation error
            return False

    def convert(self, data_spec, format):
        """Convert to a  compatible data format.

        :param dict data_spec: Data specification
        :param str format: The target data format
        :returns: dict

        >>> spec = {'name': 'a', 'type': 'number', 'format': 'number'}
        >>> port = Port(spec)

        >>> new_spec = port.convert({'format': 'number', 'data': 1}, 'json')
        >>> new_spec['format']
        'json'
        >>> port.fetch(new_spec)
        1
        """
        return convert(self.type, data_spec, {'format': format})

    def fetch(self, data_spec):
        """Return the data described by the given specification.

        :param dict data_spec: A data specification object
        :returns: data
        :raises ValidationError: when the validation check fails

        >>> port = Port({'name': 'a', 'type': 'number', 'format': 'number'})
        >>> port.fetch({'format': 'number', 'data': -1})
        -1
        """
        if self.auto_validate and not self.validate(data_spec):
            raise ValidationError(self, data_spec)

        if self.auto_convert:
            _data = self.convert(data_spec, self.format)
            data = _data.get('data')
        elif self.format == data_spec.get('format'):
            # TODO: This doesn't look right...
            if 'data' in self and self['data'] is not None:
                data = self['data']
            else:
                data = io.fetch(data_spec, task_input=self).get('data')
        else:
            raise Exception('Expected matching data formats ({} != {})' % (
                str(data_spec['format']), str(self.format)
            ))

        return data

    def push(self, data_spec):
        """Write data a to remote destination according the to specification.

        :param dict data_spec: A data specification object
        :returns: dict

        >>> port = Port({'name': 'a', 'type': 'number', 'format': 'number'})
        >>> port.push({'format': 'json', 'mode': 'inline', 'data': '2'})['data']
        2

        >>> port.push({'format': 'number', 'mode': 'inline', 'data': 3})['data']
        3
        """
        _spec = data_spec

        if self.auto_validate and not self.validate(_spec):
            raise ValidationError(self, _spec)

        if self.auto_convert:
            _spec = self.convert(_spec, self.format)
        elif _spec['format'] == self.format:

            data = data_spec.get('script_data')  # Is this always a task output?

            io.push(data, _spec, task_output=self.spec)
        else:
            raise Exception('Expected matching data formats ({} != {})' % (
                str(_spec['format']), str(self.format)
            ))
        return _spec


Port.make_property('name', 'The name of the port')
Port.make_property('type', 'The data type of the port', 'python')
Port.make_property('format', 'The data format of the port', 'object')
Port.make_property('auto_convert', 'If the data format is automatically', True)
Port.make_property('auto_validate', 'If the data is validated by default', True)

__all__ = (
    'Port',
    'ValidationError'
)



import collections

# Attempt to keep column ordering if objects happen to have ordered keys
field_map = collections.OrderedDict()
rows = []


def subkeys(path, obj, row):
    if isinstance(obj, dict):
        for k in obj:
            if isinstance(k, (str, unicode)):
                subkeys(path + [k], obj[k], row)
    elif len(path) > 0:
        field = '.'.join(path)
        field_map[field] = True
        row[field] = obj


for obj in input:
    row = {}
    subkeys([], obj, row)
    rows.append(row)

fields = [key for key in field_map]

output = {'fields': fields, 'rows': rows}



import core
from girder_worker.utils import JobStatus
from .app import app


def _cleanup(*args, **kwargs):
    core.events.trigger('cleanup')


@app.task(name='girder_worker.run', bind=True, after_return=_cleanup)
def run(_celery_task, *pargs, **kwargs):
    kwargs['_job_manager'] = _celery_task.job_manager \
        if hasattr(_celery_task, 'job_manager') else None

    kwargs['_celery_task'] = _celery_task

    kwargs['status'] = JobStatus.RUNNING

    return core.run(*pargs, **kwargs)



import json
import networkx as nx

input = json.loads(input)

nodes = [data for data in input if data['type'] == 'node']
edges = [data for data in input if data['type'] == 'link']
edge_tuples = [(e['source']['$oid'], e['target']['$oid']) for e in edges]

# If there is more than one edge with the same src/target, it's a multigraph
is_multigraph = len(edge_tuples) > len(set(edge_tuples))

if is_multigraph:
    output = nx.MultiDiGraph()
else:
    output = nx.DiGraph()

for node in nodes:
    output.add_node(node['_id']['$oid'], node['data'] if 'data' in node else {})

for edge in edges:
    output.add_edge(edge['source']['$oid'],
                    edge['target']['$oid'],
                    attr_dict=edge['data'] if 'data' in edge else {})
